#!/usr/bin/env python3
import getpass
import json
from pathlib import Path
import time
from typing import Any, Dict, List, Optional, Tuple, Union

import requests
import teselagen

DEFAULT_HOST_URL: str = "https://platform.teselagen.com"
DEFAULT_API_TOKEN_NAME: str = "x-tg-cli-token"


def load_from_json(filepath: Path) -> Any:
    """

    Loads a JSON file.

    Args:
        filepath (Path) : Path to the input JSON.

    Returns:
        (Any) : It returns a JSON object.

    """
    absolute_path: Path = filepath.absolute()

    json_obj: Any = json.loads(absolute_path.read_text())

    return json_obj


def get_project_root() -> Path:
    """ Returns project's root folder <absolute/path/to>/lib
    """
    return Path(teselagen.__path__[0]).parent.resolve()


def get_credentials_path() -> Path:
    """ Returns path to where credentials file should be
    """
    return get_project_root() / '.credentials'


def get_test_configuration_path() -> Path:
    """ Returns path to where .test_configuration file should be
    """
    return get_project_root() / '.test_configuration'


## CLIENT UTILS


def get_credentials(
    username: Optional[str] = None,
    password: Optional[str] = None,
) -> Tuple[str, str]:
    """

        It prompts the user for credentials in case username/password aren't provided
        and credentials file wasn't found.

        Args:
            username (Optional[str]) :  A valid username address to authenticate.
                If not provided, it will be prompted.

                Default : None

            password (Optional[str]) : A password to authenticate with. If not
                provided it will be prompted.

                Default : None

        Returns:
            (Tuple[str, str]) : It returns the credentials as a tuple of strings,
                containing the username and password.

                (user, password)
    """
    # Check if crentials are defined on a file
    file_credentials = load_credentials_from_file()
    username = file_credentials[0] if username is None else username
    password = file_credentials[1] if password is None else password
    # If credentials aren't defined, get them from user input
    try:
        username = input(f"Enter username: ") if username is None else username
        password = getpass.getpass(prompt=f"Password for {username}: "
                                  ) if password is None else password
    except IOError as e:
        msg = ("""There was an error with user input. If you are making parallel
               tests, make sure you are avoiding 'input' by adding CREDENTIALS
               file.""")
        raise IOError(msg)
    # End
    return username, password


def load_credentials_from_file(
    path_to_credentials_file: str = None
) -> Tuple[Optional[str], Optional[str]]:
    """Load credentials from json credentials file

    The credentials file should contain a JSON object
    with the following keys (and the values)

    ```
    {
        "username": "user",
        "password": "password"
    }
    ```

    Args:
        path_to_credentials_file (str): Path to the file. If not set it will check for `.credentials` file
            at the folder that holds this method.
    Returns:
        username, password: Username and password strings if info is found in a credentials file, and (None, None)
            if not.
    """
    if path_to_credentials_file is None:
        path_to_credentials_file = str(get_credentials_path())
    if not Path(path_to_credentials_file).is_file():
        return None, None
    credentials: Dict = load_from_json(filepath=Path(path_to_credentials_file))
    return credentials['username'], credentials['password']


def handler(func):
    """

    Decorator to handle the response from a request.

    """

    def wrapper(**kwargs):
        # -> requests.Response
        if "url" not in kwargs.keys():
            message = "url MUST be specified as keyword argument"
            raise Exception(message)

        url: str = kwargs.pop("url")

        try:
            response: requests.Response = func(url, **kwargs)

            if response.ok:
                return response

            elif response.status_code == 400:
                resp = json.loads(response.content)
                message: str = f"{response.reason}: {resp['error']}"
                raise Exception(message)

            elif response.status_code == 401:
                message: str = f"URL : {url} access is unauthorized."
                raise Exception(message)

            elif response.status_code == 404:
                message: str = f"URL : {url} cannot be found."
                raise Exception(message)

            elif response.status_code == 405:
                message: str = f"Method not allowed. URL : {url}"
                raise Exception(message)

            # TODO : Add more exceptions.

            else:
                # reason: str = response.reason
                message: str = f"Got code : {response.status_code}. Reason : {response.reason}"
                raise Exception(message)

        except Exception as e:
            raise

    return wrapper


def parser(func):
    """

    Decorator to parse the response from a request.

    """

    def wrapper(**kwargs) -> Dict[str, Union[str, bool, None]]:

        if "url" not in kwargs.keys():
            message = "url MUST be specified as keyword argument"
            raise Exception(message)

        url: str = kwargs["url"]

        response: requests.Response = func(**kwargs)

        # TODO : Should we get/return JSON Serializables values ?
        # status 204 has no content.
        if response.status_code == 204:
            print("Deletion successful.")
            return {}

        response_as_json: Dict[str, Union[str, bool, None]] = {
            "url": url,
            "status": response.ok,
            "content": response.content.decode() if response.ok else None
        }

        return response_as_json

    return wrapper


def requires_login(func):
    """ Decorator to perform login beforehand, if necessary

    Add this decorator to any function from Client or a children
    that requires to be logged in.
    """

    def wrapper(self, *args, **kwargs):
        if self.auth_token is None:
            self.login()
            if self.auth_token is None:
                raise Exception(
                    "Could not access API, access token missing. Please use the 'login' function to obtain access."
                )
        return func(self, *args, **kwargs)

    return wrapper


@parser
@handler
def get(url: str, params: dict = None, **kwargs):
    """

    Same arguments and behavior as requests.get but handles exceptions and
    returns a dictionary instead of a requests.Response.

    NOTE : url key MUST be passed in arguments.

    Returns:
        (Dict[str, Union[str, bool, None]]) : It returns a dictionary with the
            following keys and value types:

            {   "url" : str,
                "status" : bool,
                "content" : Optional[str, None]
            }

    Raises:

        (Exception) : It raises an exception if something goes wrong.

    """
    response: requests.Response = requests.get(url, params=params, **kwargs)
    return response


@parser
@handler
def post(url: str, **kwargs) -> requests.Response:
    """

    Same as requests.post but handles exceptions and returns a dictionary
    instead of a requests.Response.

    NOTE : url key MUST be passed in arguments.

    Example :

        url = "https://www.some_url.com/"
        response = post(url=url)

    Wrong usage:

        url = "https://www.some_url.com/"
        response = post(url)

    Returns:

        (Dict[str, Union[str, bool, None]]) : It returns a dictionary with the
            following keys and value types:

            {   "url" : str,
                "status" : bool,
                "content" : Optional[str, None]
            }

    Raises:

        (Exception) : It raises an exception if something goes wrong.

    """
    response: requests.Response = requests.post(url, **kwargs)
    return response


@parser
@handler
def delete(url: str, **kwargs) -> requests.Response:
    """
    Same as requests.delete but handles exceptions and returns a dictionary
    instead of a requests.Response.

    """
    response: requests.Response = requests.delete(url, **kwargs)
    return response


@parser
@handler
def put(url: str, **kwargs):
    response: requests.Response = requests.put(url, **kwargs, timeout=None)
    return response


def download_file(url: str, local_filename: str = None, **kwargs) -> str:
    """ Downloads a file from the specified url
    """
    if local_filename is None:
        local_filename = url.split('/')[-1]
    # NOTE the stream=True parameter below
    chunk_size = None
    with requests.get(url, stream=True, **kwargs) as r:
        r.raise_for_status()
        with open(local_filename, 'wb') as f:
            for chunk in r.iter_content(chunk_size=chunk_size):
                # If you have chunk encoded response uncomment if
                # and set chunk_size parameter to None.
                if chunk:
                    f.write(chunk)
    return local_filename



#!/usr/local/bin/python3
# Copyright (C) 2018 TeselaGen Biotechnology, Inc.
# License: MIT

import getpass
import json
from os.path import join
from typing import Any, cast, Dict, List, Optional, Tuple, Union
from urllib.parse import urlencode
from urllib.parse import urljoin
from urllib.parse import urlparse

import numpy as np
import pandas as pd
from teselagen.utils import DEFAULT_API_TOKEN_NAME
from teselagen.utils import DEFAULT_HOST_URL
from teselagen.utils import download_file
from teselagen.utils import get
from teselagen.utils import post
from teselagen.utils import put
from teselagen.utils import requires_login

SUPPORTED_AA_EXPORT_FORMATS = ['JSON', 'FASTA', 'GENBANK']
# NOTE : Related to Postman and Python requests
#       "body" goes into the "json" argument
#       "Query Params" goes into "params" argument


class DESIGNClient():
    ALLOWED_SEQ_FORMATS = {'json', 'fasta', 'genbank'}
    URL_GET_ASSEMBLY_REPORT = "/assembly-report/export"

    def __init__(self, teselagen_client: Any):
        module_name: str = "design"

        self.host_url = teselagen_client.host_url
        self.headers = teselagen_client.headers
        # Here we define the Base CLI URL.
        api_url_base: str = f"{self.host_url}/{module_name}/cli-api"
        # EXPORT
        # GET
        # /export/sequence/{format}/{sequenceId}
        self.export_dna_sequence_url: str = f"{api_url_base}/export/sequence/"
        # GET
        # /export/sequences/{format}/
        self.export_dna_sequences_url: str = f"{api_url_base}/export/sequences/json"
        # GET
        # /export/aminoacids/{format}/{sequenceId}
        #self.export_aminoacid_sequence_url: str = f"{api_url_base}/export/aminoacids"

        ## IMPORT
        # POST
        #self.import_dna_sequence_url: str = f"{api_url_base}/import/sequence"
        # POST
        #self.import_aminoacid_sequence_url: str = f"{api_url_base}/import/aminoacids"
        ## DESIGN
        # GET
        # /designs/{id}
        self.get_design_url: str = f"{api_url_base}/designs"
        # DEL
        # /designs/{id}
        #self.delete_design_url: str = f"{api_url_base}/designs"
        # GET
        # /designs
        self.get_designs_url: str = f"{api_url_base}/designs"
        # POST
        # /designs
        self.post_designs_url: str = f"{api_url_base}/designs"
        # POST
        # /codon-optimization-jobs
        self.post_codon_op: str = f"{api_url_base}/codon-optimization-jobs"
        # GET
        # /codon-optimization-jobs
        self.get_codon_op_result: str = f"{api_url_base}/codon-optimization-jobs"

        # GET
        # /codon-optimization-jobs
        self.get_assembly_report_url: str = f"{api_url_base}/assembly-report/export" + "/{}"

        ## RBS Calculator API Tesealgen Integration endpoints

        # GET
        # /rbs-calculator/status
        self.rbs_calculator_status_url: str = f"{api_url_base}/rbs-calculator/status"

        # POST
        # /rbs-calculator/submit
        self.rbs_calculator_submit_url: str = f"{api_url_base}/rbs-calculator/submit"

        # GET
        # /rbs-calculator/jobs/
        self.rbs_calculator_jobs_url: str = join(api_url_base,
                                                 "rbs-calculator/jobs")

        # GET
        # /rbs-calculator/jobs/:jobId
        self.rbs_calculator_job_url: str = join(api_url_base,
                                                "rbs-calculator/jobs") + "/{}"

        # GET
        # /rbs-calculator/organisms
        self.rbs_calculator_organisms_url: str = f"{api_url_base}/rbs-calculator/organisms"

        # POST
        # /import/aminoacids
        self.import_aa_url: str = f"{api_url_base}/import/aminoacids"

        # GET
        # /export/aminoacids/:format/:sequenceId
        # The aa export url accepts to url params:
        #  - the first one being the format ("FASTA", "GENBANK", "JSON")
        #  - the second one being the sequence id.
        # example: `self.export_aa_url.format("JSON", 2)`
        self.export_aa_url: str = join(api_url_base, "export",
                                       "aminoacids") + "/{}/{}"

    def get_dna_sequence(
            self,
            seq_id: int,
            out_format: str = 'json',
            out_filepath: Optional[str] = None) -> Union[str, dict]:
        """Gets full sequence record from its ID

        Args:
            seq_id (int): Sequence id (can be found within the sequence detail url at UI)
            out_format (str, optional): Output format. Use 'json' and the method
                will return a dict and 'fasta', or 'genbank' to return a string based
                on those formats. This also determines the output format when writing an
                output file. Defaults to 'json'.
            out_filepath (Optional[str], optional): Path to output file. If None it will
                not create any file. Defaults to None.

        Raises:
            ValueError: If format not available

        Returns:
            Union[str, dict]: A dict object with json data or a string if another format
                is chosen.
        """
        if out_format not in self.ALLOWED_SEQ_FORMATS:
            raise ValueError(
                f"Format {out_format} not in {self.ALLOWED_SEQ_FORMATS}")
        url = urljoin(self.export_dna_sequence_url, f'{out_format}/{seq_id}')
        response: Dict[str, Any] = get(url=url, headers=self.headers)
        # Write output file
        if out_filepath is not None:
            with open(out_filepath, 'w') as f:
                f.write(response["content"])
        # Parse json
        if out_format == 'json':
            response['content'] = json.loads(response["content"])
        # Finish
        return response["content"]

    def get_dna_sequences(self, name: str) -> List[dict]:
        """ Get all sequences which names matches a string

        Args:
            name (str): name of sequences to download

        Returns:
            List[dict]: [description]
        """
        args = {'name': name}
        response: Dict[str, Any] = get(url=self.export_dna_sequences_url,
                                       headers=self.headers,
                                       params=args)
        out = json.loads(response["content"])
        return out

    def get_designs(self,
                    name: Optional[str] = None,
                    gql_filter: Optional[dict] = None) -> List[dict]:
        """Retrieves a list of designs summary

        Args:
            name (str, optional):  Design's name
                to filter query. Defaults to None.
            gql_filter (dict, optional): GraphicQL
                filter dictionary. May be used to add additional filter conditions.
                See api-docs.Defaults to None.

        Returns:
            List[dict]: A list of designs info. Each dict in the list contains name and
                id of each design.
        """
        # Prepare parameters
        args: Dict[str, Any] = {'gqlFilter': {}}
        if name is not None:
            args['gqlFilter']["name"] = name
        if gql_filter is not None:
            args['gqlFilter'].update(gql_filter)
        # Param gqlFilter should be a json string
        args['gqlFilter'] = json.dumps(args['gqlFilter'])
        # Make request and process output
        response: Dict[str, Any] = get(url=self.get_designs_url,
                                       headers=self.headers,
                                       params=args)
        out = json.loads(response["content"])
        # Remove unuseful key
        for el in out:
            el.pop("__typename")
        return out

    def get_design(self, design_id: Union[int, str]) -> dict:
        """ Retrieves the design with specified id

        Raises error if design_id is not found

        Args:
            design_id (str, int):  Design's id

        Returns:
            dict: A dict containing designs information
        """
        response: Dict[str, Any] = get(url=f"{self.get_design_url}/{design_id}",
                                       headers=self.headers)
        #params=args)
        out = json.loads(response["content"])
        return out

    def post_design(self, design: dict, allow_duplicates: bool = False):
        """ Sumbits a new design into DESIGN module

        Args:
            design (dict): A dictionary with the design. This dictionary
                is very complex, but it can be generated easily with the
                `build_design_from_candidates` method at *utils*
            allow_duplicates (bool): Set to True to avoid raising errors on
                detection of parts duplication (default value is False).

        Returns:
            dict: On success, returns a dict containing the id of the new design (ex: `{'id': 5}` )
        """
        body = {"designJson": design, "allowDuplicates": allow_duplicates}
        response: Dict[str, Any] = post(url=self.post_designs_url,
                                        headers=self.headers,
                                        json=body)
        return json.loads(response["content"])

    def get_assembly_report(self, report_id: int, local_filename=None) -> str:
        """
        Retrieves an assembly report given an id

        Args:
            report_id (int): The id of report as can be seen
                on the browser URL for that report view in the
                DESIGN module
        Returns:
            str path to output file

        """
        if local_filename is None:
            local_filename = f"report_{report_id}.zip"
        # url = f"{self.api_url_base}{self.URL_GET_ASSEMBLY_REPORT}/{report_id}"
        url = self.get_assembly_report_url.format(report_id)
        return download_file(url=url,
                             local_filename=local_filename,
                             headers=self.headers)

    def post_codon_optimization_job(self,
                                    algorithm="ALGORITHMS_NAME",
                                    parameters={}):
        body = {"algorithm": algorithm, "parameters": parameters}
        response: Dict[str, Any] = post(url=self.post_codon_op,
                                        headers=self.headers,
                                        json=body)
        return json.loads(response["content"])

    def get_codon_optimization_job_results(self, job_id):
        response: Dict[str,
                       Any] = get(url=f"{self.get_codon_op_result}/{job_id}",
                                  headers=self.headers)
        #params=args)
        out = json.loads(response["content"])
        return out

    ## RBS Calculator Methods.

    def rbs_calculator_set_token(self, rbs_token: str = None) -> None:
        """
        Sets TeselaGen-RBS calculator integration token.

        Args:
            rbs_token(str): Integration token. This is required to consume Tesealgen/RBS Calculator API. Please ask TeselaGen for your integration token.

        Returns:
            dict: {authenticated: boolean, success: boolean}
        """
        rbs_token = getpass.getpass(
            prompt="Enter x-tg-rbs-token: ") if rbs_token is None else rbs_token
        self.headers = {**self.headers, "x-tg-rbs-token": rbs_token}

    def rbs_calculator_status(self) -> dict:
        """
        Checks the status of the RBS Calculator Integration API.

        Returns:
            dict: {authenticated: boolean, success: boolean}
        """
        try:
            result: Dict[str, Any] = get(url=self.rbs_calculator_status_url,
                                         headers=self.headers)
        except Exception as e:
            return {"error": e}

        return result["content"]

    def rbs_calculator_get_jobs(self, job_id: str = None) -> dict:
        """
        Fetches an RBS Calculator Job with the provided job_id.

        Args:
            job_id (str): ID of an RBS Calculator Job
        Returns:
            dict: {authenticated: boolean, success: boolean}
        """

        try:
            result: Dict[str, Any] = get(
                url=self.rbs_calculator_job_url.format(job_id)
                if job_id is not None else self.rbs_calculator_jobs_url,
                headers=self.headers)
        except Exception as e:
            return {"error": e}

        return result["content"]

    def rbs_calculator_organisms(
            self,
            as_dataframe: bool = True) -> Union[pd.DataFrame, Dict[str, Any]]:
        """
        Fetches all available organisms or host supported by the RBS Calculator tools.

        Args:
            as_dataframe(bool): Whether to return the response as a dataframe.
        Returns:
            List[Dict[str, str]]: A list of all the available organisms/hosts with their names and NCBI Accession IDs.
        """

        try:
            result: Union[pd.DataFrame, Dict[str, Any]] = get(url=self.rbs_calculator_organisms_url,
                                         headers=self.headers)
        except Exception as e:
            return {"error": e}

        result = json.loads(result["content"])
        result = pd.DataFrame(result) if as_dataframe else result
        return result

    def rbs_calculator_submit_job(self, algorithm: str, params: Dict[str, Any]) -> dict:
        """
        Submits a job to the RBS Calculator API Version v2.1. For deeper information on the RBS Calculator tools please refer to the following documentation:

        - Paper: https://www.researchgate.net/publication/51155303_The_Ribosome_Binding_Site_Calculator.
        - Browser Application: https://salislab.net/software/
        - Swagger API Documentation: https://app.swaggerhub.com/apis-docs/DeNovoDNA/JobControl/1.0.1


        The TeselaGen/RBS Integration currently supports one of the three following RBS Calculator Tools:

        - "ReverseRBS": Calls the RBS Calculator in Reverse Engineering mode to predict the translation
            initiation rate of each start codon in a mRNA sequence. ([Predict Translation Rates](https://salislab.net/software/predict_rbs_calculator))

            parameters:
                mRNA (str): Valid 'GATCU' mRNA sequence.
                long_UTR (boolean): Enables long UTRs.
                organism (str): Valid organism name. (for all available organism names, please call the 'rbs_calculator_organisms' function)


        - "RBSLibraryCalculator_SearchMode": Calls the RBS Library Calculator in Search mode to design a ribosome binding site library
            to maximally cover a selected  translation rate space between a targeted minimum and maximum rate
            using the fewest number of RBS variants ([Optimize Expression Levels](https://salislab.net/software/design_rbs_library_calculator)).

            parameters:
                CDS (str): Valid 'GATCU' coding sequence.
                RBS_Constrains (str): Either an empty string or a valid degenerate nucleotide sequence ('GATCURYSWKMBDHVN').
                initial_RBS_sequence (str): Either an empty string or a valid 'GATCU' RBS sequence.
                    This is used to initialize ther RBS sequence exploration algorithm. If an empty string is provided,
                    a random RBS sequence will be used as the initilizing sequence.
                library_size (int): Number of RBS sequences in your library.
                maximum_consecutive_degeneracy (int): The maximum number of consecutive degeneracy nucleotides for the RBS library designs.
                minimum_translation_initiation_rate (int): Lowest translation rate desired for your RBS library (proportional scale varies from 1 to 1,000,000).
                maximum_translation_initiation_rate (int): Highest translation rate desired for your RBS library (proportional scale varies from 1 to 1,000,000).
                organism (str): Valid organism name. (for all available organism names, please call the 'rbs_calculator_organisms' function).
                pre_sequence (str): Either an empty string or a valid 'GATCU' mRNA sequence that is required to appear upstream (5') of the RBS sequence.


        - "RBSLibraryCalculator_GenomeSearchMode": Calls the RBS Library Calculator in Genome Editing mode to design a genomic ribosome binding site library
            to maximally cover a selected translation rate space between a targeted minimum and maximum rate,  while introducing the
            fewest number of consecutive genomic mutations. ([Optimize Expression Levels](https://salislab.net/software/design_rbs_library_calculator)).

            parameters:
                CDS (str): Valid 'GATCU' coding sequence.
                RBS_Constrains (str): Either an empty string or a valid degenerate nucleotide sequence ('GATCURYSWKMBDHVN').
                genomic_RBS_sequence (str): Genomic RBS sequence. Must be a valid 'GATCU' sequence.
                initial_RBS_sequence (str): Either an empty string or a valid 'GATCU' RBS sequence.
                    This is used to initialize ther RBS sequence exploration algorithm. If an empty string is provided,
                    a random RBS sequence will be used as the initilizing sequence.
                library_size (int): Number of RBS sequences in your library.
                maximum_consecutive_degeneracy (int): The maximum number of consecutive degeneracy nucleotides for the RBS library designs.
                minimum_translation_initiation_rate (int): Lowest translation rate desired for your RBS library (proportional scale varies from 1 to 1,000,000).
                maximum_translation_initiation_rate (int): Highest translation rate desired for your RBS library (proportional scale varies from 1 to 1,000,000).
                organism (str): Valid organism name. (for all available organism names, please call the 'rbs_calculator_organisms' function).
                pre_sequence (str): Either an empty string or a valid 'GATCU' mRNA sequence that is required to appear upstream (5') of the RBS sequence.


        Args:
            algorithm (str): This should be one for the three algorithm described above currently suppoprted by the TeselaGen/RBS Integration.
            params (dict): These are the parameters required by the chosen algorithms according to the RBS Calculator API Swagger specifications mentioned above.
                        For more information on the parameters meaning refer to the https://salislab.net/software/ browser application.

                        Examples for the tools parameter inputs are as folows:

                        'ReverseRBS' params:
                            {
                                "mRNA": "YOUR_mRNA_ SEQUENCE",
                                "long_UTR": false,
                                "organism": "Acetobacter pomorum"
                            }

                        'RBSLibraryCalculator_SearchMode' params:
                            {
                                "CDS": "YOUR_CDS_SEQUENCE",
                                "RBS_Constraints": 'TCTAGANNNNNNNNNNNNNNNNNNNNNNNNNGAATTC',
                                "initial_RBS_sequence": "GATTGCGTGTGAGTTCTGGCACGGAGGAGCACGTA",
                                "library_size": 16,
                                "maximum_consecutive_degeneracy": 6,
                                "maximum_translation_initiation_rate": 100,
                                "minimum_translation_initiation_rate": 10,
                                "organism": "Escherichia coli str. K-12 substr. MG1655",
                                "pre_sequence": ""
                            }
                        'RBSLibraryCalculator_GenomeSearchMode' params:
                            {
                                "CDS": "YOUR_CDS_SEQUENCE",
                                "RBS_Constraints": "",
                                "genomic_RBS_sequence": "CUCGUACGGUGCUAACGUGCUUAGU",
                                "initial_RBS_sequence": "",
                                "library_size": 16,
                                "maximum_consecutive_degeneracy": 6,
                                "maximum_translation_initiation_rate": 100,
                                "minimum_translation_initiation_rate": 10,
                                "organism": "Escherichia coli str. K-12 substr. MG1655",
                                "pre_sequence": ""
                            }

        Returns:
            JSON with RBS Calculator job response. This may depend on the chosen tool.
        """

        _params: str = json.dumps({**params, **{"algorithm": algorithm}})
        try:
            result: Dict[str, Any] = post(url=self.rbs_calculator_submit_url,
                          data=_params,
                          headers=self.headers)
        except Exception as e:
            return {"error": e}

        result = json.loads(result["content"])
        return result

    ## Amino acid sequence Methods.
    def import_aa_sequences(self,
                            aa_sequences: Union[pd.DataFrame, List[List[str]],
                                                List[Tuple[str, str]],
                                                List[Dict[str, str]]],
                            tags: Optional[List[Dict[str, int]]] = None):
        '''
        This function imports one or many amino acid sequences by means of TeselaGen's DESIGN API.

        Args:
            aa_sequences(Union[pd.DataFrame, List[Dict[str,str]], List[Tuple[str, str]]): Amino acid sequences data. The data can come in three different ways:
                - as a pandas dataframe with 2 columns. Where the first column contains the sequence names and the second column contains the amino acid sequence string.
                - as a list of python dictionaries, where each dictionary is of the form `{"AA_NAME": SEQUENCE_NAME, "AA_SEQUENCE": SEQUENCE_STRING}`.
                - as a list of 2-element tuples, where the firt element is the sequence name and the second element the sequence string.

            tags(Optional[List[int]]): A list of integer tag IDs with which each amino acid sequence will be tagged with.
                (NOTE: tags cannot be created on-the-fly through this function, it only accepts tag IDs that are already created in the DESIGN Module).

        Returns:
            A JSON object with the following two key/values:
                - createdAminoAcidSequences(): 'id' and 'name' of the created amino acid sequences.
                - existingAminoAcidSequences(): 'id' of the updated amino acid sequences.
        '''
        params = {}
        if (aa_sequences is not None):
            if isinstance(aa_sequences, pd.DataFrame):
                params['name'] = aa_sequences.iloc[:, 0].values.tolist()
                params['contents'] = aa_sequences.iloc[:, 1].values.tolist()

            elif isinstance(aa_sequences, list):
                if all(
                        isinstance(x, list) and len(x) == 2
                        for x in aa_sequences):
                    params['name'] = list(map(lambda x: x[0], aa_sequences))
                    params['contents'] = list(map(lambda x: x[1], aa_sequences))

                if all(
                        isinstance(x, tuple) and len(x) == 2
                        for x in aa_sequences):
                    params['name'] = list(map(lambda x: x[0], aa_sequences))
                    params['contents'] = list(map(lambda x: x[1], aa_sequences))

                elif all(isinstance(x, dict) for x in aa_sequences):
                    params['name'] = list(
                        map(lambda x: x['AA_NAME'], aa_sequences))
                    params['contents'] = list(
                        map(lambda x: x['AA_SEQUENCE'], aa_sequences))

                else:
                    raise ValueError(
                        "All elements in list argument 'aa_sequences' must either be 2-element tuples or properly formatted dictionaries accroding to the function's Args description."
                    )
            else:
                raise ValueError(
                    f"Type {type(aa_sequences)} for argument 'aa_sequences' is not supported."
                )

        else:
            raise Exception("The 'aa_sequences' argument is mandatory.")

        if (tags is not None and isinstance(tags, list)):
            params["tags"] = list(map(lambda x: {"id": x}, tags))

        try:
            result: Dict[str, Any] = post(url=self.import_aa_url,
                          data=json.dumps(params),
                          headers=self.headers)
        except Exception as e:
            return e

        parsed_api_result = json.loads(result["content"])

        formatted_response = {}

        created_aa_seqs_key = 'createdAminoAcidSequences'
        updated_aa_seqs_key = 'existingAminoAcidSequences'

        if (created_aa_seqs_key in parsed_api_result.keys() and
                len(parsed_api_result[created_aa_seqs_key]) > 0):
            formatted_response[created_aa_seqs_key] = list(
                map(lambda x: {
                    "id": x['id'],
                    "name": x["name"]
                }, parsed_api_result[created_aa_seqs_key]))

        if (updated_aa_seqs_key in parsed_api_result.keys() and
                len(parsed_api_result[updated_aa_seqs_key]) > 0):
            formatted_response[updated_aa_seqs_key] = list(
                map(lambda x: {"id": x['id']},
                    parsed_api_result[updated_aa_seqs_key]))

        return formatted_response

    def export_aa_sequence(self, aa_sequence_id: int, format: str = "JSON"):
        '''
        This functions exports one amino acid sequence from TeselaGen DESIGN Module. It requires the TeselaGen amino acid sequence ID.

        Args:
            aa_sequence_id(int): This is an integer ID corresponding to the TeselaGen amino acid sequence ID.

            format(str): This is the format in which the amino acid sequence will be parsed into.
                Available formats are:
                    - JSON (teselagen specific)
                    - FASTA
                    - GENBANK

        Returns:
            (Any): Amino acid sequence information depending on the format chosen. The 'JSON' format will provide the following properties:

                - id: Amino acid sequence DESIGN ID.
                - name: Amino acid sequence name.
                - size: Number of residues for the amino acid sequence.
                - molecularWeight: Teselagen calculated molecular weight of the amino acid sequence.
                - extinctioCoefficient: Teselagen calculated extinction coefficient of the amino acid sequence.
                - proteinSequence: String with the amino acid sequence.
                - createdAt: Date in which the amino acid sequence record was created in TeselaGen.
                - createdAt: Date in which the amino acid sequence record was last updated in TeselaGen.
                - regionAnnotations: Teselagen region annotations (optional).
                - isoPoint: Sequence isoelectric point (optional).
                - uniprotId: UniProt ID for the amino acid sequence (optional).
                - tags: any TeselaGen tags with which the amino acid sequence has been tagged (optional).
        '''
        if not isinstance(aa_sequence_id, int):
            raise ValueError(
                f"Argument 'aa_sequence_id' must be of type 'int', but received type '{type(aa_sequence_id)}'."
            )
        if format not in SUPPORTED_AA_EXPORT_FORMATS:
            raise ValueError(
                f"Argument 'format' can only be one of this three strings: JSON, FASTA or GENBANK."
            )
        try:
            result: Dict[str, Any] = get(url=self.export_aa_url.format(format, aa_sequence_id),
                         headers=self.headers)
        except Exception as e:
            return e

        if format == 'JSON':
            parsed_response = json.loads(result['content'])
            formatted_response = {
                "id":
                    parsed_response["id"],
                "name":
                    parsed_response['name'],
                'isoPoint':
                    parsed_response['isoPoint'],
                'uniprotId':
                    parsed_response['uniprotId'],
                'size':
                    parsed_response['size'],
                'molecularWeight':
                    parsed_response['molecularWeight'],
                'extinctionCoefficient':
                    parsed_response['extinctionCoefficient'],
                'proteinSequence':
                    parsed_response['proteinSequence'],
                'regionAnnotations':
                    parsed_response['regionAnnotations'],
                'tags':
                    list(
                        map(
                            lambda x: {
                                'id': x['tag']['id'],
                                'name': x['tag']['name']
                            }, parsed_response['taggedItems'])),
                'createdAt':
                    parsed_response['createdAt'],
                'updatedAt':
                    parsed_response['updatedAt'],
            }
        else:
            formatted_response = result['content']

        return formatted_response

    def export_aa_sequences(self,
                            aa_sequence_ids: Union[int, np.ndarray, List[int]],
                            format: str = "JSON"):
        '''
        This functions exports one or many amino acid sequences from TeselaGen DESIGN Module. It requires one or a list of DESIGN amino acid sequence IDs.

        Args:
            aa_sequence_ids(Union[np.ndarray, List[int]]): This can be either a single integer DESIGN amino acid sequence ID or a list of them.

            format(str): This is the format in which the amino acid sequence will be parsed into.
                Available formats are:
                    - JSON (TeselaGen specific)
                    - FASTA
                    - GENBANK

        Returns:
            Returns:
            (List[Any]): A list of amino acid sequence information depending on the format chosen. The 'JSON' format will provide the following keys:

                - id: Amino acid sequence DESIGN ID.
                - name: Amino acid sequence name.
                - size: Number of residues for the amino acid sequence.
                - molecularWeight: Teselagen calculated molecular weight of the amino acid sequence.
                - extinctioCoefficient: Teselagen calculated extinction coefficient of the amino acid sequence.
                - proteinSequence: String with the amino acid sequence.
                - createdAt: Date in which the amino acid sequence record was created in TeselaGen.
                - createdAt: Date in which the amino acid sequence record was last updated in TeselaGen.
                - regionAnnotations: Teselagen region annotations (optional).
                - isoPoint: Sequence isoelectric point (optional).
                - uniprotId: UniProt ID for the amino acid sequence (optional).
                - tags: any TeselaGen tags with which the amino acid sequence has been tagged (optional).
        '''

        _sequence_ids: List[int] = []
        if isinstance(aa_sequence_ids, int):
            _sequence_ids.append(aa_sequence_ids)
        if isinstance(aa_sequence_ids, np.ndarray):
            if all(isinstance(x, int) for x in cast(np.ndarray, aa_sequence_ids)):
                _sequence_ids.extend(cast(np.ndarray, aa_sequence_ids).tolist())
            else:
                raise ValueError(
                    "All elements in list argument 'aa_sequence_ids' must be of type int."
                )
        if isinstance(aa_sequence_ids, list):
            if all(isinstance(x, int) for x in aa_sequence_ids):
                _sequence_ids.extend(aa_sequence_ids)
            else:
                raise ValueError(
                    "All elements in list argument 'aa_sequence_ids' must be of type int."
                )
        else:
            raise ValueError(
                "Argument 'aa_sequence_ids' must either be of type int, List[int] or numpy array of int elements."
            )

        #TODO: Optimize exporting multiple amino acid sequences by extending DESIGN API such that implements an endpoint supporting this.

        formatted_response = []
        for sequence_id in _sequence_ids:
            aa_sequence = self.export_aa_sequence(aa_sequence_id=sequence_id,
                                                  format=format)
            formatted_response.append(aa_sequence)

        return formatted_response



#!/usr/local/bin/python3
# Copyright (C) 2020 TeselaGen Biotechnology, Inc.
# License: MIT

import argparse
from teselagen.api.design_client import DESIGNClient

def get_report(report_id: int, host_url: str="https://platform.teselagen.com", local_filename: str=None)->str:
    """Downloads an assembly report from DESIGN module

    Args:
        report_id (int): An integer with the id of the report (this id can be seen from the address field
            when accessing the report from the browser)
        host_url (str): Address of your favorite TG platform instance, ex: "platform.teselagen.com"
        local_filename (str): Name of the output file that will contain the downloaded zipfile, ex: "report_0608.zip"
            By default it will download the file with a new name that contains the id

    Return:
        local_filename (str): Contains the name of the output file
        """
    design_client = DESIGNClient(
        host_url=host_url
    )
    return design_client.get_assembly_report(
        report_id=report_id,
        local_filename=local_filename)

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Gets a report from DESIGN platform')
    parser.add_argument(
        '--id',  type=int, required=True,
        dest='report_id',
        help='id of the report')
    parser.add_argument(
        '--url',  type=str,
        dest='host_url',
        help='The url of the hosting platform. Example: "platform.teselagen.com"')
    parser.add_argument(
        '--file',  type=str,
        dest='local_filename',
        help='The filepath of the downloaded file')
    args = parser.parse_args()
    # Remove None parameters
    kwargs = {k:v for k,v in args.__dict__.items() if v is not None}
    res = get_report(**kwargs)
    print(f"Report saved to '{res}'")


#!/usr/local/bin/python3
# Copyright (C) 2018 TeselaGen Biotechnology, Inc.
# License: MIT

import json
from os.path import join
from typing import Any, Dict, List, Optional, Tuple, TypeVar, Union

import numpy as np
import pandas as pd
import requests
from teselagen.utils import DEFAULT_API_TOKEN_NAME
from teselagen.utils import DEFAULT_HOST_URL
from teselagen.utils import get
from teselagen.utils import post
from teselagen.utils import put
from teselagen.utils import requires_login

# NOTE : Related to Postman and Python requests
#       "body" goes into the "json" argument
#       "Query Params" goes into "params" argument

ALLOWED_MODEL_TYPES: List[Union[str, None]] = [
    "predictive", "evolutive", "generative", None
]

PREDICTIVE_MODEL = 'predictive'
EVOLUTIVE_MODEL = 'evolutive'
GENERATIVE_MODEL = 'generative'
GENERATIVE_MODEL_DATA_SCHEMA = [{
    'id': 0,
    'name': 'sequence',
    'type': 'target',
    'value_type': 'aa-sequence'
}]


class DISCOVERClient():

    def __init__(self, teselagen_client: Any):
        module_name: str = "evolve"  #(now the 'discover' module)

        self.host_url = teselagen_client.host_url
        self.headers = teselagen_client.headers
        # Here we define the Base CLI URL.
        api_url_base: str = f"{self.host_url}/{module_name}/cli-api"
        # Here we define the client endpoints
        # Example :
        #    self.some_endpoint_url: str = f"{self.api_url_base}/some_endpoint"
        self.create_model_url: str = f"{api_url_base}/create-model"

        self.get_model_url: str = f"{api_url_base}/get-model"
        self.get_models_by_type_url: str = f"{api_url_base}/get-models-by-type"
        self.get_model_datapoints_url: str = f"{api_url_base}/get-model-datapoints"

        self.submit_model_url: str = f"{api_url_base}/submit-model"
        self.submit_multi_objective_optimization_url: str = f"{api_url_base}/multi-objective-optimization-task"
        self.get_multi_objective_optimization_url: str = f"{api_url_base}/multi-objective-optimization-task/" + "/{}"
        self.delete_model_url: str = f"{api_url_base}/delete-model"
        self.cancel_model_url: str = f"{api_url_base}/cancel-model"
        self.cancel_task_url: str = f"{api_url_base}/cancel-task" + "/{}"

        self.get_task_url: str = f"{api_url_base}/get-tasks" + "/{}"
        self.get_models_url: str = f"{api_url_base}/get-models"
        self.get_completed_tasks_url: str = f"{api_url_base}/get-completed-tasks"

        self.crispr_guide_rnas_url: str = f"{api_url_base}/crispr-grnas"

    def _get_data_from_content(self, content_dict: dict) -> dict:
        """Checks that an output dict from evolve endpoint is healthy, and returns the 'data' field

        Args:
            content_dict (dict): content field (as dictionary) from an api endpoint response

        Raises:
            IOError: If dictionary isn't healthy

        Returns:
            dict: data field from endpoint response
        """
        if content_dict["message"] != "Submission success.":
            message = content_dict["message"]
            raise IOError(f"A problem occured with query: {message}")
        if 'data' not in content_dict:
            raise IOError(f"Can`t found 'data' key in response: {content_dict}")
        return content_dict["data"]

    def get_model_info(self, model_id: int):
        """ Retrieves model general information

        This will return a JSON object with the metadata of a model filtered
        by the provided model ID.

        Args :
            model_id (int) :
                Model identifier.

        Returns :
            () : A dict containing model info. An example is shown below:

        ```
        {
            "id": "0",
            "labId": "1",
            "modelType": "predictive",
            "name": "My First Predictive Model",
            "description": "This is an example model",
            "status": "completed-successfully",
            "evolveModelInfo": {
                "microserviceQueueId":
                "1",
                "dataSchema": [{
                    "id": "1",
                    "name": "Descriptor1",
                    "value_type": "numeric",
                    "type": "descriptor"
                }, {
                    "id": "1",
                    "name": "Descriptor2",
                    "value_type": "numeric",
                    "type": "descriptor"
                }, {
                    "id": "2",
                    "name": "Target",
                    "value_type": "numeric",
                    "type": "target"
                }],
                "modelStats": {
                    "MAE": 45
                }
            }
        }
        ```

        """
        body = {"id": str(model_id)}
        response: Dict[str, Any] = post(url=self.get_model_url,
                                        headers=self.headers,
                                        json=body)
        response["content"] = json.loads(response["content"])
        # Check output
        return self._get_data_from_content(response["content"])

    def get_models_by_type(self, model_type: Optional[str] = None):
        """

        This will return a JSON object with the metadata of multiple models,
        filtered by the provided `model_type`.

        Args :
            model_type (str) :

        ```
            "predictive"
            "evolutive"
            "generative"
             None
        ```

        Returns :
            () :

        ```

        {
            "message":
            "Submission success.",
            "data": [{
                "id": "1",
                "labId": "1",
                "modelType": "evolutive",
                "name": "My First Evolutive Model",
                "description": "This is an example model",
                "status": "completed-successfully",
                "evolveModelInfo": {
                    "microserviceQueueId":
                    "1",
                    "dataSchema": [{
                        "id": "1",
                        "name": "Descriptor1",
                        "value_type": "numeric",
                        "type": "descriptor"
                    }, {
                        "id": "1",
                        "name": "Descriptor2",
                        "value_type": "numeric",
                        "type": "descriptor"
                    }, {
                        "id": "2",
                        "name": "Target",
                        "value_type": "numeric",
                        "type": "target"
                    }],
                    "modelStats": {
                        "MAE": 45
                    }
                }
            }, {
                "id": "2",
                "labId": "1",
                "modelType": "evolutive",
                "name": "My Second Evolutive Model",
                "description": "This is an example model",
                "status": "completed-successfully",
                "evolveModelInfo": {
                    "microserviceQueueId":
                    "1",
                    "dataSchema": [{
                        "id": "1",
                        "name": "Descriptor1",
                        "value_type": "numeric",
                        "type": "descriptor"
                    }, {
                        "id": "1",
                        "name": "Descriptor2",
                        "value_type": "numeric",
                        "type": "descriptor"
                    }, {
                        "id": "2",
                        "name": "Target",
                        "value_type": "numeric",
                        "type": "target"
                    }],
                    "modelStats": {
                        "MAE": 40
                    }
                }
            }]
        }

        ```

        """
        if model_type not in ALLOWED_MODEL_TYPES:
            raise ValueError(f"Type: {model_type} not in {ALLOWED_MODEL_TYPES}")
        # body = {"modelType": "null" if model_type is None else model_type}
        body = {"modelType": model_type}
        response: Dict[str, Any] = post(url=self.get_models_by_type_url,
                                        headers=self.headers,
                                        json=body)
        response["content"] = json.loads(response["content"])
        return self._get_data_from_content(response["content"])

    def get_model_datapoints(
        self,
        model_id: Union[int, str],
        datapoint_type: str,
        batch_size: int,
        batch_number: int,
    ) -> Dict[str, Any]:
        """
            This will return a JSON object with an array of datapoints filtered by
            the provided model ID and datapoint type. This array will come in the
            data field in the response body. Each element of the array has a
            datapoint field, this corresponds to a JSON object with the datapoint
            data.

            Args :

                model_id (int) :
                    ID of the model

                datapoint_type (str) :
                    The `datapoint_type` has two options :

                        "input"
                        "output"

                    One can fetch only input datapoints (a.k.a training datapoints)
                    or just fetch the output datapoint (a.k.a predicted datapoints
                    not seen in the training dataset).

                batch_size (int) :
                    `batch_size` refers to the number of datapoints to fetch from
                    the database table.

                batch_number (int) :
                    `batch_number` depends on `batch_size`, and determines the
                    index position offset of length `batch_size` from where to
                    start fetching datapoints.

            Returns :
                - An object with a 'data' key with the list of datapoints along with their predictions.

            ```
                {
                    "message": "Submission success.",
                    "data": [{ ... }, { ... }, { ... }]
                }
            ```
        """

        body = {
            "modelId": str(model_id),
            "datapointType": datapoint_type,
            "batchSize": batch_size,
            "batchNumber": batch_number
        }

        response: Dict[str, Any] = post(
            url=self.get_model_datapoints_url,
            headers=self.headers,
            json=body,
        )

        responseContent: Dict[str, Any] = json.loads(response["content"])

        datapoints: List[Dict[str, Any]] = []
        if 'data' in responseContent:
            datapoints = [{
                key: value
                for key, value in element['datapoint'].items()
                if key != 'set_tag' and "PCA" not in key
            }
                          for element in responseContent['data']]

        responseContent.update({'data': datapoints})

        return responseContent

    def submit_model(
        self,
        data_input: List[Any],
        data_schema: List[Any],
        model_type: str,
        configs: Optional[Any] = None,
        name: str = "",
        description: Optional[str] = None,
    ) -> Dict[str, Any]:
        """

        Submits a model for training.

        Args :
            data_input (List[Any]) :
                This is required and must contain a JSON array of JSON objects
                with the input training data.

                These objects must be consistent with the `data_schema`
                property.

        ```
                [{
                    "Descriptor1": "A0",
                    "Descriptor2": "B1",
                    "Target": "1"
                }, {
                    "Descriptor1": "A0",
                    "Descriptor2": "B2",
                    "Target": "2"
                }, {
                    "Descriptor1": "A0",
                    "Descriptor2": "B3",
                    "Target": "3"
                }]
        ```

            data_schema (List[Dict[str, Any]]) :
                This is an array of the schema of the input data columns.

                The `name` property corresponds to the column's name.

                The `type` property determines whether the column is a "target"
                or a "descriptor" (feature). Only "target" and "descriptor"
                are supported.

                The `value_type` type determines the type of the column's
                values. Only "numeric" and "categoric" are supported.

        ```
                [{
                    "id": "1",
                    "name": "Descriptor1",
                    "value_type": "categoric",
                    "type": "descriptor"
                }, {
                    "id": "2",
                    "name": "Descriptor2",
                    "value_type": "categoric",
                    "type": "descriptor"
                }, {
                    "id": "3",
                    "name": "Target",
                    "value_type": "numeric",
                    "type": "target"
                }]
        ```
                - `id` : corresponds to the id (position) of the column in the
                    dataset.
                - `name` : corresponds to the name of the column (descriptor
                    or target)
                - `type` : describes whether the field is a descriptor
                    (feature) or a target.
                - `value_type` : defines the type of value of this column.
                    Available types are "numeric" or "categoric".

            model_type (str) :
                The type of model wanting to submit.

                Either "predictive", "evolutive" or "generative".

                NOTE : If submitting a "generative" model, there's no "descriptor"
                       column, in fact there should only be one "target"
                       column with the amino acid sequence. This needs to be properly set in the dataSchema
                       field according to the documentation.

            configs (Optional[Any]) :
                This is an advanced property containing advanced configuration
                for the training execution. Please refer to Teselagen's Data
                Science Team.

            name (str) :
                This sets the Evolve Model's name.

            description (Optional[str]) :
                This gives the Evolve Model's a description.

        Returns :
            (dict) : A dictionary containing info of the submitted job. En example is shown below:

            ```
            {
                "authToken": "1d140371-a59f-4ad2-b57c-6fc8e0a20ff8",
                "checkInInterval": null,
                "controlToken": null,
                "id": "36",
                "input": {
                    "job": "modeling-tool",
                    "kwargs": {}
                },
                "lastCheckIn": null,
                "missedCheckInCount": null,
                "result": null,
                "resultStatus": null,
                "service": "ds-tools",
                "serviceUrl": null,
                "startedOn": null,
                "status": "created",
                "taskId": null,
                "trackingId": null,
                "completedOn": null,
                "createdAt": "2020-10-29T13:18:06.167Z",
                "updatedAt": "2020-10-29T13:18:06.271Z",
                "cid": null,
                "__typename": "microserviceQueue"
            }
            ```

        """
        body = {
            "dataInput": data_input,
            "dataSchema": data_schema,
            "modelType": model_type,
            "configs": {} if configs is None else configs,
            "name": name,
            "description": "" if description is None else description
        }
        response: Dict[str, Any] = post(url=self.submit_model_url,
                                        headers=self.headers,
                                        json=body)

        response["content"] = json.loads(response["content"])
        return self._get_data_from_content(response["content"])

    def submit_prediction_task(
        self,
        data_input: List[Dict[str, Any]],
        data_schema: List[Dict[str, Any]],
        model_id: Union[int, str],
    ) -> Dict[str, Any]:
        """
            Submits a task used to run predictions on a ist of datapoints using a pre-trained Predictive Model.

            Args:
                - data_input (List[Dict[str, Any]]): Datapoints in the same format described in the submit_model function.
                - data_schema (List[Dict[str, Any]]): Data schema in the same format described in the submit_model function.
                - model_id (Union[int, str]): ID of the pre-trained predictive model going to be used to run predictions for the datapoints in the data_input list.

            Returns:
                - A Task object with metadata information on the submitted task including its ID for later retrieval.
        """
        body = {
            "dataInput": data_input,
            "dataSchema": data_schema,
            "modelType": "predictive",
            "predictiveModelId": model_id,
            # "configs": {} if configs is None else configs,
            "name": "pretrained",
            # "description": "" if description is None else description
        }
        response: Dict[str, Any] = post(
            url=self.submit_model_url,
            headers=self.headers,
            json=body,
        )

        responseContent: Dict[str, Any] = json.loads(response["content"])

        responseContent['data'].update({'pretrainedModelId': model_id})

        return responseContent

    def get_prediction_task(
        self,
        task: Any,
        batch_size: int,
        batch_number: int = 0,
    ) -> Dict[str, Any]:
        """
            Returns the result of a prediction task. If the task is running it will return a task status object.
            If the task finished it will return the list of datapoints with their prediction.

            Args:
                - task (Any): The task object obtaine from submitting a prediction task.
                - batch_size (int): Number of datapoints to be fetched.
                - batch_number (int): When providing a batch_size, the full set of datapoints is divided by batches of size batch_size.
                    batch_number is used to tell which batch of batch_size datapoints to fetch. Defaults to the first batch (batch_number=0).

            Returns:
                - Depending on the status of the submitted task, a task object with task metadata information including its task status wil be returned.
                    If the task is complete, it will return the set of datapoints along with their predictions.
        """
        model_id = task['data']['modelId']
        task_id = task['data']['id']
        task_response: Dict[str, Any] = self.get_task(task_id=task_id)
        task_status: Dict[str, Any] = task_response['data'][0]['status']

        results: Dict[str, Any] = {}
        if task_status == 'completed-successfully':
            results = self.get_model_datapoints(
                model_id=model_id,
                datapoint_type='output',
                batch_size=batch_size,
                batch_number=batch_number,
            )
        else:
            results = task_response['data'][0]

        return results

    def submit_multi_objective_optimization(
        self,
        data_input: List[Any],
        data_schema: List[Any],
        pretrainedModelIds: List[Union[int, str]] = None,
        configs: Optional[Any] = None,
    ):
        """
            Submits a multi objective optimization task.

            Args :
                data_input (List[Any]) :
                    This is required and must contain a JSON array of JSON objects
                    with the input training data.

                    These objects must be consistent with the `data_schema`
                    property.

            ```
                    [{
                        "Descriptor1": "A0",
                        "Descriptor2": "B1",
                        "Target_1": "1",
                        "Target_2": "-1"
                    }, {
                        "Descriptor1": "A0",
                        "Descriptor2": "B2",
                        "Target_1": "2",
                        "Target_2": "-2"
                    }, {
                        "Descriptor1": "A0",
                        "Descriptor2": "B3",
                        "Target_1": "3",
                        "Target_2": "-3"
                    }]
            ```

                data_schema (List[Any]) :
                    This is an array of the schema of the input data columns.

                    The `name` property corresponds to the column's name.

                    The `type` property determines whether the column is a "target"
                    or a "descriptor" (feature). Only "target" and "descriptor"
                    are supported.

                    The `value_type` type determines the type of the column's
                    values. Only "numeric" and "categoric" are supported.

            ```
                    [{
                        "name": "Descriptor1",
                        "value_type": "categoric",
                        "type": "descriptor"
                    }, {
                        "name": "Descriptor2",
                        "value_type": "categoric",
                        "type": "descriptor"
                    }, {
                        "name": "Target_1",
                        "value_type": "numeric",
                        "type": "target"
                    }, {
                        "name": "Target_2",
                        "value_type": "numeric",
                        "type": "target"
                    }]
            ```
                    - `name` : corresponds to the name of the column (descriptor
                        or target)
                    - `type` : describes whether the field is a descriptor
                        (feature) or a target.
                    - `value_type` : defines the type of value of this column.
                        Available types are "numeric" or "categoric"


                configs (Optional[Any]) :
                    This is an advanced property containing advanced configuration
                    for the training execution. Please refer to Teselagen's Data
                    Science Team.

            Returns :
                (dict) : A dictionary containing info of the submitted job. En example is shown below:

                ```
                {
                    "authToken": "1d140371-a59f-4ad2-b57c-6fc8e0a20ff8",
                    "checkInInterval": null,
                    "controlToken": null,
                    "id": "36",
                    "input": {
                        "job": "modeling-tool",
                        "kwargs": {}
                    },
                    "lastCheckIn": null,
                    "missedCheckInCount": null,
                    "result": null,
                    "resultStatus": null,
                    "service": "ds-tools",
                    "serviceUrl": null,
                    "startedOn": null,
                    "status": "created",
                    "taskId": null,
                    "trackingId": null,
                    "completedOn": null,
                    "createdAt": "2020-10-29T13:18:06.167Z",
                    "updatedAt": "2020-10-29T13:18:06.271Z",
                    "cid": null,
                    "__typename": "microserviceQueue"
                }
                ```
        """
        body = {
            "dataInput": data_input,
            "dataSchema": data_schema,
            "predictiveModelIds": pretrainedModelIds,
            "configs": {} if configs is None else configs,
        }
        response: Dict[str, Any] = post(
            url=self.submit_multi_objective_optimization_url,
            headers=self.headers,
            json=body,
        )
        response["content"] = json.loads(response["content"])
        return response["content"]

    def get_multi_objective_optimization(self, taskId: str) -> Any:

        response: Dict[str, Any] = get(
            url=self.get_multi_objective_optimization_url.format(taskId),
            headers=self.headers)

        response["content"] = json.loads(response["content"])
        return response["content"]

    def delete_model(self, model_id: int):
        """

        Deletes a model matching the specified `model_id`.

        Args :
            model_id (int) :
                The model id that wants to be deleted.

        Returns :
            () :

        """
        body = {"id": str(model_id)}
        response: Dict[str, Any] = post(url=self.delete_model_url,
                                        headers=self.headers,
                                        json=body)
        response["content"] = json.loads(response["content"])
        return self._get_data_from_content(response["content"])
        # raise NotImplementedError

    def cancel_model(self, model_id: int):
        """

        Cancels the submission of a model matching the specified `model_id`.

        Args :
            model_id (int) :
                The model id that wants to be cancelled.

        Returns :
            () :

        """
        body = {"id": str(model_id)}
        response: Dict[str, Any] = post(url=self.cancel_model_url,
                                        headers=self.headers,
                                        json=body)
        response["content"] = json.loads(response["content"])
        return self._get_data_from_content(response["content"])

    def get_task(self, task_id: Union[int, str]) -> Any:
        """
            returns the status of a task based on the Task ID.

            Args :
                task_id (int) :
                    The task id that wants to be cancelled.

            Returns :
                - A task object with task metadata including its ID and status.
        """
        response: Dict[str, Any] = get(
            url=self.get_task_url.format(task_id),
            headers=self.headers,
        )

        return json.loads(response["content"])

    def cancel_task(self, task_id: int) -> Any:
        """

            Cancels the submission of a task matching the specified `task_id`.

            Args :
                task_id (int) :
                    The task id that wants to be cancelled.

            Returns :
                () :

        """
        response: Dict[str, Any] = post(
            url=self.cancel_task_url.format(task_id),
            headers=self.headers,
        )
        return json.loads(response["content"])

    def design_crispr_grnas(self,
                            sequence: str,
                            target_indexes: Optional[Tuple[int, int]] = None,
                            target_sequence: Optional[str] = None,
                            pam_site: str = 'NGG',
                            min_score: float = 40.0,
                            max_number: Optional[int] = 50):
        body: Dict[str, Any] = {
            'data': {
                'sequence': sequence
            },
            'options': {
                'pamSite': pam_site,
                'minScore': min_score
            }
        }
        if target_indexes is not None:
            body['data']['targetStart'] = target_indexes[0]
            body['data']['targetEnd'] = target_indexes[1]
        if target_sequence is not None:
            body['data']['targetSequence'] = target_sequence
        if max_number is not None:
            body['options']['maxNumber'] = max_number
        response: Dict[str, Any] = post(url=self.crispr_guide_rnas_url,
                                        headers=self.headers,
                                        json=body)
        return json.loads(response["content"])

    def submit_generative_model(
            self,
            aa_sequences: Optional[Union[np.ndarray, List[str]]] = None,
            aa_sequence_ids: Optional[Union[np.ndarray, List[int]]] = None,
            model_name: Optional[
                str] = 'Unnamed Generative Model (Python Package)',
            model_description: Optional[str] = None,
            model_configs: Optional[dict] = {}) -> Dict[str, Any]:
        """
        Calls DISCOVER API 'POST /submit-model' endpoint to train an amino acid sequence Generative Model.

        Args:
            aa_sequences(Optional[List[str]]): List of strings corresponding to valid amino acid sequences.
                Currently, generative models only support training sequences of 10 to 50 amino acids. Only IUPAC 20 amino acids are supported.

            aa_sequence_ids(Optional[List[int]]): List of amino acid sequence IDs. These IDs correspond to TeselaGen's DESIGN Module IDs.
                These IDs are returned by the 'DESIGNClient.import_aa_sequences(...)' function when importing new or existant aa sequences.
                But you can also obtain your amino acid sequence IDs via the DESIGN Module Web Browser App from the 'Molecules > Amino Acid Sequences' Library viewer.

            model_name(Optional[str]): String as an optional name for your model. Default name is going to be: 'Unnamed Generative Model (Python Package)'.

            model_description(Optional[str]): String as an optional description for your model.

            model_configs(Optional[dict]): This is an advanced property containing advanced configuration
                for the training execution. Please refer to Teselagen's Data Science Team.

        Returns:
            (dict) : A Python Dictionary with information about the model submission, inlcuding the task id used to check the status of the training.

            ```
            {
                "id": "36",
                "lastCheckIn": null,
                "result": null,
                "status": "created",
                "completedOn": null,
                "createdAt": "2020-10-29T13:18:06.167Z",
                "updatedAt": "2020-10-29T13:18:06.271Z",
            }
            ```

        """

        kwargs: Dict[str, Any] = {
            'data_schema': GENERATIVE_MODEL_DATA_SCHEMA,
            'model_type': GENERATIVE_MODEL,
            'name': model_name,
            'description': model_description,
            'configs': model_configs
        }

        if (aa_sequences is not None):
            if isinstance(aa_sequences, list) or isinstance(
                    aa_sequences, np.ndarray):
                if all(isinstance(x, str) for x in aa_sequences):
                    kwargs['data_input'] = list(
                        map(lambda x: {'sequence': x}, aa_sequences))
                else:
                    raise ValueError(
                        "All amino acid sequences must be of type string.")
        elif (aa_sequence_ids is not None):
            if isinstance(aa_sequence_ids, list) or isinstance(
                    aa_sequence_ids, np.ndarray):
                if all(isinstance(x, int) for x in aa_sequence_ids):
                    NotImplementedError(
                        "Passing sequence IDs is not yet supported.")
                    # TODO: import sequences from DESIGN using the IDs in aa_sequence_ids.
                    # exported_sequences = DESIGNClient.export_aa_sequences(...)
                    # kwargs['data_input'] = list(map(lambda x: {'sequence': x}, exported_sequences))

        # response = self.submit_model(**kwargs)
        response: Dict[str, Any] = self.submit_model(
            data_input=kwargs['data_input'],
            data_schema=kwargs['data_schema'],
            model_type=kwargs['model_type'],
            configs=kwargs['configs'],
            name=kwargs['name'],
            description=kwargs['description'])

        formatted_response = {
            # When submitting a model a new microservice job is created with ID=response['id']
            "jobId": response['id'],
            # When submitting a model a new model record is created with ID=response['modelId']
            "modelId": response['modelId'],
            "status": response['status'],
            "createdAt": response['createdAt'],
            "updatedAt": response['updatedAt'],
        }
        return formatted_response



#!/usr/local/bin/python3
# Copyright (C) 2020 TeselaGen Biotechnology, Inc.
# License: MIT
import math
import json
import textwrap
from pathlib import Path
from typing import List
import argparse
import sys

import pandas as pd
"""
This module provides a command line tool that generates a json design
that can be imported into TeselaGen's DESIGN module. As an input, this
tool uses a csv data table as the one can be exported from the *evolutions*
algorithm's result, at EVOLVE. Some usage examples can be found below
"""

def build_design_from_candidates(
    candidates_data: List[dict],
    bin_cols: List[str],
    #lab_id: str,
    #user_id: str,
    name: str,
    priority_col='Priority')->dict:
    ''' Builds design dict from candidates data

    Args:
        candidates_data: Records object from EVOLVE output data csv
        bin_cols: List containing the names of the bins to be read from data
        name: Name of design as will be seen on DESIGN.

    Returns:
        A dict that contains design information. This dict can be serialized as
        json to be used within the DESIGN module.

    '''
    designs = {'design':{}}
    # Use just the elements with valid priority (suggested candidates)
    data = [d for d in candidates_data if not math.isnan(d[priority_col])]
    print(f"Generating design using {len(data)} candidates")
    # Create bins
    designs['design']['bin'] = {f"{i+1}":{'id':f"{i+1}",
                                          'direction': True,
                                          'name': name,
                                          'iconId':"1"} for i,name in enumerate(bin_cols)}
    # Create Bincards
    designs['design']['binCard'] = {f"{i+1}":{'id': f"{i+1}",
                                          'index': i,
                                          'cardId': "1",
                                          'binId': f"{i+1}"} for i, name in enumerate(bin_cols)}
    # Create card
    designs['design']['card'] = {"1": {"id": "1",
                                       "isRoot": True,
                                       "circular": True,
                                       "name": "Target Construct"}}
    # Create design
    designs['design']['design'] = {"1":{"id": "1",
                                        #"userId": user_id,
                                        #"labId": lab_id,
                                        "name": name,
                                        "type": "grand-design",
                                        "layoutType": "list",
                                        "numRows": len(data)}}
    # Create elements
    designs['design']['element'] = {}
    el_i = 0
    for i_bin, bin_name in enumerate(bin_cols):
        for i_el, el in enumerate(data):
            el_i += 1
            designs['design']['element'][f"{el_i}"] = {
            	"id": f"{el_i}",
                "name": el[bin_name],
                "index": i_el,
                "binId": f"{i_bin+1}"}
    # Set icon
    designs['design']['icon'] = {"1": {"id": "1",
                                 "isSbol": True,
                                 "name": "USER-DEFINED",
                                 "path": "M 5 40 L 5 60 L 45 60 L 45 40 Z"}}
    # Maybe this is not needed
    designs["hdeDesignExportVersion"] = 3
    #print("Done!")
    return designs


def candidates_to_design(input_file: str, bin_cols: List[str], name: str, output: str):
    ''' Read candidates file (csv) and generates json file that can be read from DESIGN
    module.

    Args:
        input_file: Path to the input csv file as string.
        bin_cols: A list with the names of the bin columns at the csv file
        name: Name of output design
        output: Path of the output file as string.
    '''
    # Get arguments
    if 'csv' in input_file:
        # Loads data from a csv file downloaded from EVOLVE UI
        data_df = pd.read_csv(input_file)
        data = data_df.to_dict(orient="records")
    # Build dictionary object with design
    design_dict = build_design_from_candidates(
        candidates_data=data,#: List[dict],
        bin_cols= bin_cols,
        name= name)
    # Write design to disc
    #Path('_export/').mkdir(parents=True, exist_ok=True)
    with open(output, 'w') as f:
        json.dump(design_dict, f)
        print(f"Created design at '{output}'")


def args_parser(args_in):
    """Parses arguments from the command line tool"""
    parser = argparse.ArgumentParser(
        prog='candidates_to_design',
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=textwrap.dedent('''\
            Script for generating a design json from a set of evolution's candidates.

            Example usage:
                Download output data from evoultion result "Teselagen Example Evolutive Model".
                Store this file in this same folder with the name "evolve_exported_results.csv".
                The next line will create a file with the default name "EVOLVE_candidates.json"
                with a design containing valid candidates.

                >>> python3 candidates_to_design.py --bin_cols "Teselagen Enzyme A" "Teselagen Enzyme B" --input "evolve_exported_results.csv"

            '''))
    parser.add_argument('--input',
                        default="input.csv",
                        #metavar='-i',
                        type=str,
                        help='Path to csv input file',
                        )
    parser.add_argument('--bin_cols',
                        #metavar='-i',
                        required=True,
                        type=str,
                        nargs='+',
                        help='List containing the names of the bins to be read from data. These names should be column names at the input',
                        )
    parser.add_argument('--name',
                        default="EVOLVE_candidates",
                        #metavar='-i',
                        type=str,
                        help='Path to csv input file',
                        )
    parser.add_argument('--output',
                        default="EVOLVE_candidates.json",
                        #metavar='-i',
                        type=str,
                        help='Path to json output file',
                        )
    return parser.parse_args(args_in)

# Command line tool, parses arguments and call candidates_to_design
if __name__ == "__main__":
    args = args_parser(sys.argv[1:])
    print(args)
    candidates_to_design(
        input_file=args.input,
        bin_cols=args.bin_cols,
        name=args.name,
        output=args.output
    )
    print("Done!")


#!/usr/bin/env python3
import json
from pathlib import Path
from typing import Any, Dict, List, Optional, Union
from itertools import product
import uuid

import pytest
import requests_mock
import fastaparser

from teselagen.api import TeselaGenClient, DISCOVERClient
from teselagen.utils import load_from_json, get_project_root

MODEL_TYPES_TO_BE_TESTED: List[Optional[str]] = [
    "predictive", "evolutive", "generative", "null"
]

@pytest.mark.incremental
class TestDISCOVERClient():
    @pytest.fixture
    def client(self, host_url, api_token_name) -> TeselaGenClient:
        """

        A TeselaGen client instance.

        Returns:
            (TeselaGenClient) : An instance of TeselaGen client.

        """
        tg_client = TeselaGenClient(api_token_name=api_token_name, host_url=host_url)
        return tg_client

    @pytest.fixture
    def logged_client(self, client: TeselaGenClient) -> DISCOVERClient:
        """

        A logged EVOLVE client instance.

        Returns:
            (DISCOVERClient) : An instance of the EVOLVE client.

        """
        expiration_time: str = "30m"
        client.login(
                     expiration_time=expiration_time)
        return client.discover

    @pytest.fixture
    def submitted_model_name(self, logged_client):
        # Define synthetic problem parameters
        params = {
            "name": f"Model X times Y {uuid.uuid1()}",
            "data_input": [{"X": str(el[0]), "Y": str(el[1]), "Z": el[0]*el[1]} for el in product(range(10), range(10))],
            "data_schema": [
                {"name": "X", "id":0, "value_type":"categoric", "type": "descriptor"},
                {"name": "Y", "id":1, "value_type":"categoric", "type": "descriptor"},
                {"name": "Z", "id":2, "value_type":"numeric", "type": "target"}],
            "model_type": "predictive"
        }
        result = logged_client.submit_model(**params)
        return params['name']

    def test_client_attributes(self, logged_client: DISCOVERClient):

        # We check if the client has the required attributes.
        assert hasattr(logged_client, "create_model_url")
        assert hasattr(logged_client, "get_model_url")
        assert hasattr(logged_client, "get_models_by_type_url")
        assert hasattr(logged_client, "get_model_datapoints_url")
        assert hasattr(logged_client, "submit_model_url")
        assert hasattr(logged_client, "delete_model_url")
        assert hasattr(logged_client, "cancel_model_url")
        assert hasattr(logged_client, "get_models_url")
        assert hasattr(logged_client, "get_completed_tasks_url")

    def test_login(self, client: DISCOVERClient, api_token_name):
        # Before login, the client has no tokens
        assert client.auth_token is None
        assert api_token_name not in client.headers.keys()

        # LOGIN
        expiration_time: str = "1d"
        client.login(expiration_time=expiration_time)

        # After login, the client has tokens
        assert isinstance(client.auth_token, str)
        assert api_token_name in client.headers.keys()
        assert isinstance(client.headers[api_token_name], str)

    @pytest.mark.skip(reason="Test not finished")
    #@pytest.mark.parametrize("model_type", MODEL_TYPES_TO_BE_TESTED)
    def test_get_models_by_type(self, logged_client: DISCOVERClient,
                                model_type: Optional[str]):

        client = logged_client
        response = client.get_models_by_type(model_type=model_type)
        assert isinstance(response, dict)

        # 1
        expected_keys: List[str] = ["message", "data"]
        assert all(
            [key if model_type != "null" else "message"][0] in response.keys()
            for key in expected_keys)

        # 2
        expected_keys: List[str] = [
            "id", "labId", "modelType", "name", "description", "status",
            "evolveModelInfo"
        ]

        if model_type != "null":

            for data in response["data"]:

                for key in expected_keys:
                    assert key in data.keys()

                    if key == "evolveModelInfo":

                        assert isinstance(data[key], dict)

                        expected_evolveModelInfokeys: List[str] = [
                            "microserviceQueueId", "dataSchema", "modelStats"
                        ]

                        assert all(k in data[key].keys()
                                   for k in expected_evolveModelInfokeys)

                    if key == "labId":
                        assert isinstance(data[key], str) or data[key] is None

                    else:
                        assert isinstance(data[key], str)

    # def test_get_model(self, logged_client: DISCOVERClient, model_id: int):
    #     client = logged_client
    #     response = client.get_model(model_id=model_id)
    #     assert isinstance(response, dict)

    @pytest.mark.skip(reason="Test is fine, but needs remote host to fix (https://github.com/TeselaGen/lims/pull/6506)")
    def test_design_crispr_grnas(self, logged_client):
        # Fasta file
        seq_filepath = get_project_root() / "teselagen/examples/dummy_organism.fasta"
        # Load file
        with open(seq_filepath) as fasta_file:
            parser = fastaparser.Reader(fasta_file)
            for seq in parser:
                fasta_seq=seq.sequence_as_string()
                break
        # Call method to be tested
        res = logged_client.design_crispr_grnas(
            sequence=fasta_seq,
            target_indexes=[500, 600],
        )
        assert isinstance(res, list)
        assert len(res) == 7

    def test_design_crispr_grnas_mock(self, logged_client, requests_mock):
        expected_url = logged_client.crispr_guide_rnas_url
        sequence = "AGTCAGGTACGGTACGGTACGGTATGGCAAAAGGACGGATGGACAGGCT"
        target_indexes = [10, 14]
        endpoint_output = [
            {"start": 10, "end": 12, "offTargetScore": 0.8, "forward": True, "pam": "CGG", "onTargetScore": 0.6}
        ]
        requests_mock.post(expected_url, json=endpoint_output)
        res = logged_client.design_crispr_grnas(
            sequence=sequence,
            target_indexes=target_indexes,
        )
        assert isinstance(res, list)
        assert res == endpoint_output

    @pytest.mark.skip(reason="Test is fine, but needs remote host to fix (https://github.com/TeselaGen/lims/pull/6506)")
    def test_get_model_submit_get_cancel_delete(self, logged_client, submitted_model_name):
        for n_attempt in range(3):
            res = logged_client.get_models_by_type(model_type="predictive")
            new_model = list(filter(lambda x: x['name']==submitted_model_name, res))
            if len(new_model) > 0: break
        assert len(new_model) == 1
        assert new_model[0]['status'] in {'in-progress', 'pending', 'completed-successfully', 'submitting'}
        res_cancel = logged_client.cancel_model(new_model[0]['id'])
        assert 'id' in res_cancel and res_cancel['id'] == new_model[0]['id']
        res_delete = logged_client.delete_model(new_model[0]['id'])
        assert 'id' in res_delete and res_delete['id'] == new_model[0]['id']

    def test_submit_model_mock(self, logged_client, requests_mock):
        expected_url = logged_client.submit_model_url
        endpoint_output = {
            "message": "Submission success.",
            "data": {'id':0}}
        requests_mock.post(expected_url, json=endpoint_output)
        # Define synthetic problem parameters
        params = {
            "name": f"Model X times Y {uuid.uuid1()}",
            "data_input": [{"X": str(el[0]), "Y": str(el[1]), "Z": el[0]*el[1]} for el in product(range(10), range(10))],
            "data_schema": [
                {"name": "X", "id":0, "value_type":"categoric", "type": "descriptor"},
                {"name": "Y", "id":1, "value_type":"categoric", "type": "descriptor"},
                {"name": "Z", "id":2, "value_type":"numeric", "type": "target"}],
            "model_type": "predictive",
            "configs": {},
            "description": ""
        }
        result = logged_client.submit_model(**params)
        assert result == endpoint_output['data']
        # Names to camel case:
        expected_params = params.copy()
        expected_params["dataInput"] = expected_params.pop("data_input")
        expected_params["dataSchema"] = expected_params.pop("data_schema")
        expected_params["modelType"] = expected_params.pop("model_type")
        assert requests_mock.last_request.json() == expected_params









#!/usr/bin/python3

import getpass
import json
import logging
from pathlib import Path
from typing import Dict

from teselagen.api.client import DEFAULT_HOST_URL
from teselagen.utils import get_credentials_path
from teselagen.utils import get_project_root
from teselagen.utils import get_test_configuration_path

logger = logging.getLogger(__name__)

# TODO : make methods to create credentials and test configurations, so they can be imported.

# Paths
root_path: Path = get_project_root()
credentials_path: Path = get_credentials_path()
configuration_path: Path = get_test_configuration_path()

if input("login credentials ? [yes|no] : ").lower() in ["y", "yes"]:

    # login
    username: str = input("username : ")
    password: str = getpass.getpass(prompt="password : ")

    credentials: Dict[str, str] = {"username": username, "password": password}

    print(f"Saving credentials to : {credentials_path.absolute()}")

    logger.info(f"Saving credentials to : {credentials_path.absolute()}")

    json.dump(credentials,
              credentials_path.open("w", encoding="utf-8"),
              indent=4)

    del username, password, credentials

if input("test configuration ? [yes|no] : ").lower() in ["y", "yes"]:

    # test configuration
    host_url: str = input(f"host_url [default: {DEFAULT_HOST_URL}] : ")
    configuration: Dict[str, str] = {
        "host_url": host_url if host_url is not "" else DEFAULT_HOST_URL
    }

    print(f"Saving test configuration to : {configuration_path.absolute()}")

    logger.info(
        f"Saving test configuration to : {configuration_path.absolute()}")

    json.dump(configuration,
              configuration_path.open("w", encoding="utf-8"),
              indent=4)

    del host_url, configuration



#!/usr/bin/env python3
from io import StringIO
import json
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import pytest
from teselagen.api import TeselaGenClient
from teselagen.api import TESTClient
from teselagen.utils import load_from_json

TEST_FILE_CONTENTS: str = r"""Line,Teselagen Example Descriptor 1,Teselagen Example Descriptor 2,Teselagen Example Target,Teselagen Example Target Metric
1,A0,B1,1,ug/mL
2,A0,B2,2,ug/mL
3,A0,B3,3,ug/mL
4,A0,B5,5,ug/mL
5,A0,B6,6,ug/mL
6,A0,B9,9,ug/mL
7,A0,B10,10,ug/mL
8,A0,B11,11,ug/mL
9,A0,B12,12,ug/mL
10,A0,B13,13,ug/mL"""

# TEST_FILE_CONTENTS: str = "Line,Teselagen Example Descriptor 1,Teselagen Example Descriptor 2,Teselagen Example Target,Teselagen Example Target Metric\n1,A0,B1,1,ug/mL\n2,A0,B2,2,ug/mL\n3,A0,B3,3,ug/mL\n4,A0,B5,5,ug/mL\n5,A0,B6,6,ug/mL\n6,A0,B9,9,ug/mL\n7,A0,B10,10,ug/mL\n8,A0,B11,11,ug/mL\n9,A0,B12,12,ug/mL\n10,A0,B13,13,ug/mL\n11,A0,B14,14,ug/mL\n12,A0,B15,15,ug/mL\n13,A0,B17,17,ug/mL\n14,A0,B18,18,ug/mL\n15,A1,B0,1,ug/mL\n16,A1,B2,3,ug/mL\n17,A1,B3,4,ug/mL\n18,A1,B5,6,ug/mL\n19,A1,B6,7,ug/mL\n20,A1,B7,8,ug/mL\n21,A1,B8,9,ug/mL\n22,A1,B11,12,ug/mL\n23,A1,B12,13,ug/mL\n24,A1,B13,14,ug/mL\n25,A1,B14,15,ug/mL\n26,A1,B15,16,ug/mL\n27,A1,B16,17,ug/mL\n28,A1,B17,18,ug/mL\n29,A1,B18,19,ug/mL\n30,A1,B19,20,ug/mL\n31,A2,B0,2,ug/mL\n32,A2,B1,3,ug/mL\n33,A2,B2,4,ug/mL\n34,A2,B3,5,ug/mL\n35,A2,B5,7,ug/mL\n36,A2,B6,8,ug/mL\n37,A2,B8,10,ug/mL\n38,A2,B9,11,ug/mL\n39,A2,B10,12,ug/mL\n40,A2,B12,14,ug/mL\n41,A2,B13,15,ug/mL\n42,A2,B14,16,ug/mL\n43,A2,B15,17,ug/mL\n44,A2,B16,18,ug/mL\n45,A2,B17,19,ug/mL\n46,A2,B18,20,ug/mL\n47,A3,B0,3,ug/mL\n48,A3,B2,5,ug/mL\n49,A3,B3,6,ug/mL\n50,A3,B4,7,ug/mL\n51,A3,B5,8,ug/mL\n52,A3,B6,9,ug/mL\n53,A3,B7,10,ug/mL\n54,A3,B9,12,ug/mL\n55,A3,B10,13,ug/mL\n56,A3,B11,14,ug/mL\n57,A3,B12,15,ug/mL\n58,A3,B13,16,ug/mL\n59,A3,B14,17,ug/mL\n60,A3,B15,18,ug/mL\n61,A3,B16,19,ug/mL\n62,A3,B17,20,ug/mL\n63,A3,B18,21,ug/mL\n64,A3,B19,22,ug/mL\n65,A4,B1,5,ug/mL\n66,A4,B2,6,ug/mL\n67,A4,B3,7,ug/mL\n68,A4,B4,8,ug/mL\n69,A4,B5,9,ug/mL\n70,A4,B6,10,ug/mL\n71,A4,B7,11,ug/mL\n72,A4,B8,12,ug/mL\n73,A4,B10,14,ug/mL\n74,A4,B11,15,ug/mL\n75,A4,B12,16,ug/mL\n76,A4,B13,17,ug/mL\n77,A4,B14,18,ug/mL\n78,A4,B16,20,ug/mL\n79,A4,B18,22,ug/mL\n80,A4,B19,23,ug/mL\n81,A5,B0,5,ug/mL\n82,A5,B1,6,ug/mL\n83,A5,B2,7,ug/mL\n84,A5,B3,8,ug/mL\n85,A5,B4,9,ug/mL\n86,A5,B6,11,ug/mL\n87,A5,B7,12,ug/mL\n88,A5,B9,14,ug/mL\n89,A5,B10,15,ug/mL\n90,A5,B11,16,ug/mL\n91,A5,B12,17,ug/mL\n92,A5,B13,18,ug/mL\n93,A5,B14,19,ug/mL\n94,A5,B15,20,ug/mL\n95,A5,B16,21,ug/mL\n96,A5,B17,22,ug/mL\n97,A5,B18,23,ug/mL\n98,A5,B19,24,ug/mL\n99,A6,B0,6,ug/mL\n100,A6,B1,7,ug/mL\n101,A6,B2,8,ug/mL\n102,A6,B6,12,ug/mL\n103,A6,B7,13,ug/mL\n104,A6,B8,14,ug/mL\n105,A6,B9,15,ug/mL\n106,A6,B10,16,ug/mL\n107,A6,B11,17,ug/mL\n108,A6,B12,18,ug/mL\n109,A6,B13,19,ug/mL\n110,A6,B14,20,ug/mL\n111,A6,B15,21,ug/mL\n112,A6,B17,23,ug/mL\n113,A6,B18,24,ug/mL\n114,A6,B19,25,ug/mL\n115,A7,B0,7,ug/mL\n116,A7,B2,9,ug/mL\n117,A7,B3,10,ug/mL\n118,A7,B4,11,ug/mL\n119,A7,B5,12,ug/mL\n120,A7,B6,13,ug/mL\n121,A7,B7,14,ug/mL\n122,A7,B9,16,ug/mL\n123,A7,B10,17,ug/mL\n124,A7,B11,18,ug/mL\n125,A7,B12,19,ug/mL\n126,A7,B14,21,ug/mL\n127,A7,B16,23,ug/mL\n128,A7,B18,25,ug/mL\n129,A7,B19,26,ug/mL\n130,A8,B0,8,ug/mL\n131,A8,B3,11,ug/mL\n132,A8,B5,13,ug/mL\n133,A8,B6,14,ug/mL\n134,A8,B7,15,ug/mL\n135,A8,B9,17,ug/mL\n136,A8,B10,18,ug/mL\n137,A8,B11,19,ug/mL\n138,A8,B12,20,ug/mL\n139,A8,B14,22,ug/mL\n140,A8,B15,23,ug/mL\n141,A8,B18,26,ug/mL\n142,A8,B19,27,ug/mL\n143,A9,B3,12,ug/mL\n144,A9,B4,13,ug/mL\n145,A9,B5,14,ug/mL\n146,A9,B6,15,ug/mL\n147,A9,B8,17,ug/mL\n148,A9,B10,19,ug/mL\n149,A9,B11,20,ug/mL\n150,A9,B12,21,ug/mL\n151,A9,B13,22,ug/mL\n152,A9,B14,23,ug/mL\n153,A9,B15,24,ug/mL\n154,A9,B16,25,ug/mL\n155,A9,B19,28,ug/mL\n156,A10,B0,10,ug/mL\n157,A10,B1,11,ug/mL\n158,A10,B3,13,ug/mL\n159,A10,B5,15,ug/mL\n160,A10,B7,17,ug/mL\n161,A10,B9,19,ug/mL\n162,A10,B10,20,ug/mL\n163,A10,B11,21,ug/mL\n164,A10,B12,22,ug/mL\n165,A10,B13,23,ug/mL\n166,A10,B14,24,ug/mL\n167,A10,B15,25,ug/mL\n168,A10,B17,27,ug/mL\n169,A10,B19,29,ug/mL\n170,A11,B0,11,ug/mL\n171,A11,B1,12,ug/mL\n172,A11,B2,13,ug/mL\n173,A11,B3,14,ug/mL\n174,A11,B4,15,ug/mL\n175,A11,B5,16,ug/mL\n176,A11,B6,17,ug/mL\n177,A11,B8,19,ug/mL\n178,A11,B10,21,ug/mL\n179,A11,B11,22,ug/mL\n180,A11,B12,23,ug/mL\n181,A11,B13,24,ug/mL\n182,A11,B15,26,ug/mL\n183,A11,B16,27,ug/mL\n184,A11,B17,28,ug/mL\n185,A11,B18,29,ug/mL\n186,A12,B0,12,ug/mL\n187,A12,B1,13,ug/mL\n188,A12,B2,14,ug/mL\n189,A12,B3,15,ug/mL\n190,A12,B5,17,ug/mL\n191,A12,B6,18,ug/mL\n192,A12,B7,19,ug/mL\n193,A12,B8,20,ug/mL\n194,A12,B9,21,ug/mL\n195,A12,B10,22,ug/mL\n196,A12,B11,23,ug/mL\n197,A12,B12,24,ug/mL\n198,A12,B14,26,ug/mL\n199,A12,B15,27,ug/mL\n200,A12,B17,29,ug/mL\n201,A12,B18,30,ug/mL\n202,A12,B19,31,ug/mL\n203,A13,B0,13,ug/mL\n204,A13,B1,14,ug/mL\n205,A13,B2,15,ug/mL\n206,A13,B3,16,ug/mL\n207,A13,B5,18,ug/mL\n208,A13,B6,19,ug/mL\n209,A13,B8,21,ug/mL\n210,A13,B9,22,ug/mL\n211,A13,B10,23,ug/mL\n212,A13,B11,24,ug/mL\n213,A13,B12,25,ug/mL\n214,A13,B15,28,ug/mL\n215,A13,B16,29,ug/mL\n216,A13,B18,31,ug/mL\n217,A13,B19,32,ug/mL\n218,A14,B1,15,ug/mL\n219,A14,B4,18,ug/mL\n220,A14,B5,19,ug/mL\n221,A14,B6,20,ug/mL\n222,A14,B9,23,ug/mL\n223,A14,B10,24,ug/mL\n224,A14,B11,25,ug/mL\n225,A14,B12,26,ug/mL\n226,A14,B13,27,ug/mL\n227,A14,B14,28,ug/mL\n228,A14,B15,29,ug/mL\n229,A14,B16,30,ug/mL\n230,A14,B17,31,ug/mL\n231,A14,B18,32,ug/mL\n232,A14,B19,33,ug/mL\n233,A15,B0,15,ug/mL\n234,A15,B1,16,ug/mL\n235,A15,B2,17,ug/mL\n236,A15,B3,18,ug/mL\n237,A15,B5,20,ug/mL\n238,A15,B6,21,ug/mL\n239,A15,B7,22,ug/mL\n240,A15,B9,24,ug/mL\n241,A15,B11,26,ug/mL\n242,A15,B12,27,ug/mL\n243,A15,B13,28,ug/mL\n244,A15,B14,29,ug/mL\n245,A15,B16,31,ug/mL\n246,A15,B17,32,ug/mL\n247,A15,B19,34,ug/mL\n248,A16,B0,16,ug/mL\n249,A16,B2,18,ug/mL\n250,A16,B3,19,ug/mL\n251,A16,B4,20,ug/mL\n252,A16,B5,21,ug/mL\n253,A16,B6,22,ug/mL\n254,A16,B7,23,ug/mL\n255,A16,B8,24,ug/mL\n256,A16,B9,25,ug/mL\n257,A16,B10,26,ug/mL\n258,A16,B11,27,ug/mL\n259,A16,B12,28,ug/mL\n260,A16,B13,29,ug/mL\n261,A16,B14,30,ug/mL\n262,A16,B15,31,ug/mL\n263,A16,B17,33,ug/mL\n264,A16,B18,34,ug/mL\n265,A16,B19,35,ug/mL\n266,A17,B0,17,ug/mL\n267,A17,B2,19,ug/mL\n268,A17,B4,21,ug/mL\n269,A17,B5,22,ug/mL\n270,A17,B6,23,ug/mL\n271,A17,B8,25,ug/mL\n272,A17,B9,26,ug/mL\n273,A17,B10,27,ug/mL\n274,A17,B11,28,ug/mL\n275,A17,B12,29,ug/mL\n276,A17,B13,30,ug/mL\n277,A17,B14,31,ug/mL\n278,A17,B15,32,ug/mL\n279,A17,B16,33,ug/mL\n280,A17,B17,34,ug/mL\n281,A17,B18,35,ug/mL\n282,A18,B0,18,ug/mL\n283,A18,B1,19,ug/mL\n284,A18,B2,20,ug/mL\n285,A18,B3,21,ug/mL\n286,A18,B4,22,ug/mL\n287,A18,B6,24,ug/mL\n288,A18,B7,25,ug/mL\n289,A18,B8,26,ug/mL\n290,A18,B9,27,ug/mL\n291,A18,B11,29,ug/mL\n292,A18,B12,30,ug/mL\n293,A18,B13,31,ug/mL\n294,A18,B14,32,ug/mL\n295,A18,B15,33,ug/mL\n296,A18,B16,34,ug/mL\n297,A18,B17,35,ug/mL\n298,A18,B18,36,ug/mL\n299,A19,B0,19,ug/mL\n300,A19,B1,20,ug/mL\n301,A19,B3,22,ug/mL\n302,A19,B4,23,ug/mL\n303,A19,B5,24,ug/mL\n304,A19,B6,25,ug/mL\n305,A19,B7,26,ug/mL\n306,A19,B9,28,ug/mL\n307,A19,B10,29,ug/mL\n308,A19,B11,30,ug/mL\n309,A19,B12,31,ug/mL\n310,A19,B13,32,ug/mL\n311,A19,B14,33,ug/mL\n312,A19,B16,35,ug/mL\n313,A19,B19,38,ug/mL"

# Default IDs for testing purposes :
#       experimentId = 1
#       parserId = 1


@pytest.mark.incremental
class TestTESTClient():

    # TODO: figure out how to run a clean up fixture.
    # @pytest.fixture(autouse=True, scope='module')
    # def clean_up(
    #     self,
    #     logged_client: TeselaGenClient,
    #     experiment: Dict[str, Any],
    #     select_laboratory,
    # ) -> None:

    #     client = logged_client.test
    #     experiment_id: str = experiment['id']
    #     client.delete_experiment(experiment_id=experiment_id)

    @pytest.fixture
    def expiration_time(self) -> str:
        _expiration_time: str = "30m"
        return _expiration_time

    @pytest.fixture
    def headers(self) -> Dict[str, str]:
        _headers: Dict[str, str] = {"Content-type": "application/json"}
        return _headers

    @pytest.fixture
    def client(self, host_url: str, api_token_name: str) -> TeselaGenClient:
        """

        A TeselaGen client instance.

        Returns:
            (TeselaGenClient) : An instance of TeselaGen client.

        """
        test_client = TeselaGenClient(api_token_name=api_token_name,
                                      host_url=host_url,
                                      module_name='test')
        return test_client

    @pytest.fixture
    def logged_client(
        self,
        client: TeselaGenClient,
        expiration_time: str,
    ) -> TeselaGenClient:
        """

        A logged TEST client instance.

        Returns:
            (TeselaGenClient) : An instance of the TEST client.

        """
        client.login(
            expiration_time=expiration_time
            #username=credentials["test_user"],
            #passwd=credentials["test_password"],
        )
        return client

    @pytest.fixture
    def lab_id(self, client: TeselaGenClient) -> int:
        """

        Get the lab id used for testing.

        Returns:
            (int) : The laboratory identifier used for testing.

        """
        available_labs = client.get_laboratories()
        _lab_id: int = available_labs[0]['id']
        # _lab_id: int = 1

        return _lab_id

    @pytest.fixture
    def select_laboratory(self, client: TeselaGenClient, lab_id: int) -> None:
        """

        A selected lab to work on.
        """
        # available_labs = client.get_laboratories()
        # lab_id: int = available_labs[0]['id']
        client.select_laboratory(lab_id=lab_id)

    @pytest.fixture
    def experiment_id(self, experiment: Dict[str, Any]) -> str:
        """

        Get the experiment id used for testing.

        Returns:
            (int) : The experiment identifier used for testing.

        """
        _experiment_id: str = experiment["id"]
        return _experiment_id

    @pytest.fixture
    def assay_id(self) -> int:
        """

        Get the assay id used for testing.

        Returns:
            (int) : The assay identifier used for testing.

        """
        _assay_id: int = 1  # None
        return _assay_id

    @pytest.fixture
    def experiment(
        self,
        logged_client: TeselaGenClient,
        select_laboratory,
    ) -> Dict[str, Any]:
        client = logged_client.test
        experiment_name: str = "Python Test Client Experiment"
        experiment: Dict[str, Any] = client.create_experiment(
            experiment_name=experiment_name)
        return experiment

    @pytest.fixture
    def assay(
        self,
        logged_client: TeselaGenClient,
        experiment: Dict[str, Any],
        select_laboratory,
    ) -> Dict[str, Any]:
        client = logged_client.test
        experiment_id: str = experiment["id"]
        assay_name: str = "Python Test Client Assay"
        # TODO : We may need to update this, probably with a parser or
        #        parser_id fixture
        parser_id: Optional[int] = None
        response: Dict[str, Any] = client.create_assay(
            experiment_id=experiment_id,
            assay_name=assay_name,
            parser_id=parser_id,
        )
        return response

    def test_class_attributes(self) -> None:

        # We check if the class inherit the parents methods.

        # parent_class_methods: List[str] = [
        #     "register", "login", "logout", "get_server_status", "create_token",
        #     "update_token", "get_api_info", "get_current_user",
        #     "get_laboratories", "select_laboratory", "unselect_laboratory"
        # ]

        # We check if the class has the required methods.

        experiment_methods: List[str] = [
            "get_experiments", "create_experiment", "delete_experiment"
        ]

        assay_methods: List[str] = [
            "get_assays", "create_assay", "delete_assay"
        ]

        file_methods: List[str] = [
            "get_files_info", "upload_file", "download_file", "delete_file"
        ]

        metadata_methods: List[str] = [
            "get_metadata", "create_metadata", "delete_metadata"
        ]

        attributes: List[str] = [
            # *parent_class_methods,
            *experiment_methods,
            *assay_methods,
            *file_methods,
            *metadata_methods,
        ]

        assert all(hasattr(TESTClient, attribute) for attribute in attributes)

    def test_instance_attributes(self, logged_client: TeselaGenClient) -> None:

        # We check if the client inherit the required parents attributes.
        # parent_class_attributes: List[str] = ["labs_url"]

        # We check if the client has the required attributes.

        experiment_attributes: List[str] = [
            "get_experiments_url", "create_experiment_url",
            "delete_experiment_url"
        ]

        assay_attributes: List[str] = [
            "get_assays_url", "get_assays_by_experiment_url",
            "create_assay_url", "delete_assay_url"
        ]

        file_attributes: List[str] = [
            "get_files_info_url", "get_file_data_url", "delete_file_url",
            "upload_file_url", "upload_file_into_assay_url"
        ]

        metadata_attributes: List[str] = [
            "get_metadata_url", "create_metadata_url", "delete_metadata_url"
        ]

        attributes: List[str] = [
            # *parent_class_attributes,
            *experiment_attributes,
            *assay_attributes,
            *file_attributes,
            *metadata_attributes,
        ]

        assert all(
            hasattr(logged_client.test, attribute) for attribute in attributes)

    def test_login(
        self,
        client: TeselaGenClient,
        expiration_time: str,
        api_token_name,
    ) -> None:
        # Before login, the client has no tokens
        assert client.auth_token is None
        assert api_token_name not in client.headers.keys()

        # LOGIN
        client.login(  #username=credentials["test_user"],
            #passwd=credentials["test_password"],
            expiration_time=expiration_time)

        # After login, the client has tokens
        assert isinstance(client.auth_token, str)
        assert api_token_name in client.headers.keys()
        assert isinstance(client.headers[api_token_name], str)

    def test_create_experiment(self, experiment: List[Dict[str, Any]]) -> None:
        response = experiment
        assert isinstance(response, dict)
        assert "id" in response.keys()
        assert response["id"] is not None

    def test_delete_experiment(
        self,
        logged_client: TeselaGenClient,
        experiment: List[Dict[str, Any]],
    ) -> None:
        client = logged_client.test
        experiment_id: str = experiment["id"]
        response = client.delete_experiment(experiment_id=experiment_id)
        assert response is None

    def test_create_assay(self, assay: List[Dict[str, Any]]) -> None:
        response = assay
        assert isinstance(response, dict)
        assert "id" in response.keys()
        assert "name" in response.keys()
        assert response["id"] is not None

    def test_delete_assay(
        self,
        logged_client: TeselaGenClient,
        lab_id: int,
        assay: Dict[str, Any],
    ) -> None:
        client = logged_client.test
        assay_id: int = int(assay["id"])
        response = client.delete_assay(assay_id=assay_id)

        assert int(response["id"]) == assay_id, "Error deleting assay."

    def test_get_experiments(
        self,
        logged_client: TeselaGenClient,
        select_laboratory,
    ) -> None:
        client = logged_client.test
        response: List[Dict[str, Any]] = client.get_experiments()

        assert isinstance(response, list)
        assert len(response) > 0
        assert all(isinstance(element, dict) for element in response)
        assert all(key in element.keys() and element[key] is not None
                   for element in response
                   for key in ["id", "name"])
        # assert all(element[key] is not None for element in response for key in ["id", "name"])

    def test_get_assays(
        self,
        logged_client: TeselaGenClient,
        experiment_id: str,
        select_laboratory,
        assay: List[Dict[str, Any]],
    ) -> None:

        client = logged_client.test

        response: List[Dict[str, Any]] = client.get_assays(
            experiment_id=experiment_id,)

        assert isinstance(response, list)
        # NOTE: It may be an empty list (?)
        assert len(response) > 0
        assert all(isinstance(element, dict) for element in response)

        # TODO : We may have to test if an "experiment" object containing an
        #       "id" and "name" is included

        assert all(key in element.keys() and element[key] is not None
                   for element in response
                   for key in ["id", "name"])
        # assert all( element[key] is not None for element in response for key in ["id", "name"])

    # @pytest.mark.skip(reason="The files endpoints are under maintenance.")
    def test_get_files(
        self,
        logged_client: TeselaGenClient,
        host_url: str,
        select_laboratory,
    ) -> None:
        if 'platform' in host_url:
            return
        client = logged_client.test
        response: List[Dict[str, Any]] = client.get_files_info()

        assert isinstance(response, list)
        assert len(response) > 0
        assert all(isinstance(element, dict) for element in response)

        # TODO : We should expect an "assay" key that may be None
        assert all(key in element.keys()
                   for element in response
                   for key in ["id", "name", "assay", "experiment"])

    # @pytest.mark.skip(reason="The files endpoints are under maintenance.")
    def test_upload_file(
        self,
        logged_client: TeselaGenClient,
        host_url: str,
        select_laboratory,
    ):
        if 'platform' in host_url:
            return
        client = logged_client.test

        filepath: Path = Path("./teselagen/api/tests/example_file.csv")
        assay_id: Optional[int] = None
        response: Dict[str, Any] = client.upload_file(
            filepath=filepath,
            assay_id=assay_id,
        )

        assert all(key in ["id", "name", "assay", "experiment"]
                   for key in response.keys())

    # @pytest.mark.skip(reason="The files endpoints are under maintenance.")
    def test_download_file(
        self,
        logged_client: TeselaGenClient,
        host_url: str,
        select_laboratory,
    ):
        if 'platform' in host_url:
            return
        client = logged_client.test
        response = client.get_files_info()
        file_id: str = response[0]['id']
        response = client.download_file(file_id=file_id)

        assert isinstance(response, StringIO)

    # @pytest.mark.skip(reason="The files endpoints are under maintenance.")
    def test_delete_file(
        self,
        logged_client: TeselaGenClient,
        host_url: str,
        select_laboratory,
    ):
        if 'platform' in host_url:
            return
        client = logged_client.test

        filepath: Path = Path("./teselagen/api/tests/example_file.csv")
        upload_response: Dict[str, Any] = client.upload_file(filepath=filepath,)

        file_id: str = upload_response['id']
        response = client.delete_file(file_id=file_id)

        assert response['id'] == file_id, "Error deleting file."

    @pytest.mark.skip(reason="We need to create a test experiment first")
    def test_download_assay(
        self,
        logged_client: TeselaGenClient,
        assay_id,
        expected_downloaded_assay_length,
        expected_downloaded_assay_row,
    ) -> None:

        client: TESTClient = logged_client.test

        # DOWNLOAD ASSAY
        full_list: Optional[bool] = None
        columns = None
        page_number: int = None
        page_size: int = None

        downloaded_assay = client.download_assay(
            assay_id=assay_id,
            full_list=full_list,
            columns=columns,
            page_number=page_number,
            page_size=page_size,
        )

        # Since we are testing this, it may not be an empty list
        assert downloaded_assay != []

        assert len(downloaded_assay) == expected_downloaded_assay_length

        row = downloaded_assay[0]
        expected_row = expected_downloaded_assay_row

        assert isinstance(row, dict)
        assert all(isinstance(key, str) for key in row.keys())
        assert all(isinstance(value, str) for value in row.values())

        assert row == expected_row

    @pytest.mark.skip(
        reason="We don't know which experiment and parser IDs to use")
    def test_upload_assays(self, logged_client) -> None:

        client = logged_client

        # UPLOAD ASSAY
        filename: str = "test_data.csv"
        contents = TEST_FILE_CONTENTS
        assay_name: str = "test_name"
        experiment_id: str = '1'
        parser_id: str = '1'

        number_of_previous_assays = len(
            client.get_assays(experiment_id=experiment_id))

        uploaded_assay_id = client.upload_assay(filename=filename,
                                                contents=contents,
                                                experiment_id=experiment_id,
                                                parser_id=parser_id,
                                                assay_name=assay_name)

        assert isinstance(uploaded_assay_id, int)

        number_of_current_assays = len(
            client.get_assays(experiment_id=experiment_id))

        assert number_of_current_assays == number_of_previous_assays + 1

        # ASSAY LENGTH
        length_of_the_uploaded_assay: int = client.get_length_of_an_assay(
            assay_id=uploaded_assay_id)

        assert length_of_the_uploaded_assay == 9



#!/usr/bin/env python3
import json
import os
from pathlib import Path
from typing import Any, Dict, List, Optional, Union
from unittest.mock import patch
import uuid

import pytest
from teselagen.api.client import DEFAULT_API_TOKEN_NAME
from teselagen.api.client import DEFAULT_HOST_URL
from teselagen.api.client import get
from teselagen.api.client import TeselaGenClient
from teselagen.utils import get_credentials_path

# module_names = ["design", "build", "test", "evolve"]
MODULES_TO_BE_TESTED: List[str] = ["test", "evolve"]


class TestTeselaGenClient:

    @pytest.fixture
    def expiration_time(self) -> str:
        _expiration_time: str = "30m"
        return _expiration_time

    @pytest.fixture
    def headers(self) -> Dict[str, str]:
        _headers: Dict[str, str] = {"Content-type": "application/json"}
        return _headers

    @pytest.fixture
    def client(
        self,
        module_name: str,
        host_url: str,
        api_token_name: str,
    ) -> TeselaGenClient:
        """

        A TesleaGenClient client instance.

        Returns:
            (TESTClient) : An instance of the TEST client.

        """

        _client = TeselaGenClient(
            module_name=module_name,
            host_url=host_url,
            api_token_name=api_token_name,
        )

        return _client

    @pytest.fixture
    def logged_client(
        self,
        client: TeselaGenClient,
        expiration_time: str,
    ) -> TeselaGenClient:
        """

        A logged TEST client instance.

        Returns:
            (TESTClient) : An instance of the TEST client.

        """
        # Test will not run without a credential file
        credentials_filepath = get_credentials_path()
        assert credentials_filepath.is_file(
        ), f"Can't found {credentials_filepath}"
        client.login(  #username=credentials["test_user"],
            #passwd=credentials["test_password"],
            expiration_time=expiration_time)
        return client

    def test_class_attributes(self) -> None:
        # Here we check if the class has the required methods.

        methods: List[str] = [
            "register", "login", "logout", "get_server_status", "create_token",
            "update_token", "get_api_info", "get_current_user",
            "get_laboratories", "select_laboratory", "unselect_laboratory"
        ]

        # static_methods: List[str] = [""]
        # attributes: List[str] = [*methods, *static_methods]

        attributes: List[str] = methods

        assert all(
            hasattr(TeselaGenClient, attribute) for attribute in attributes)

        assert isinstance(DEFAULT_API_TOKEN_NAME, str)
        assert isinstance(DEFAULT_HOST_URL, str)

    @pytest.mark.parametrize("module_name", MODULES_TO_BE_TESTED)
    def test_instance_attributes(self, client: TeselaGenClient,
                                 module_name: str) -> None:

        attributes: List[str] = [
            "host_url",
            "api_token_name",
            "register_url",
            "login_url",
            "info_url",
            "status_url",
            "auth_url",
            "labs_url",
            "headers",
            "auth_token",
        ]

        # We check if the client has the required attributes.
        assert all(hasattr(client, attribute) for attribute in attributes)

        # We verify the headers
        assert isinstance(client.headers, dict)
        assert "Content-Type" in client.headers.keys()
        assert isinstance(client.headers["Content-Type"], str)

    # There's another login test below
    # @pytest.mark.parametrize("module_name", MODULES_TO_BE_TESTED)
    # def test_03_post_to_login_endpoint(self, module_name):
    #     host_url: str = HOST_URL
    #
    #     api_url: str = f"{host_url}/{module_name}/login"
    #     headers: Dict[str, str] = {"Content-type": "application/json"}
    #     request: Dict[str, str] = {
    #         "email": credentials["test_user"],
    #         "password": credentials["test_password"]
    #     }
    #
    #     response = post(url=api_url, headers=headers, json=request)
    #     del request

    #     assert isinstance(response, dict)

    #     expected_keys: List[str] = ["content", "status", "url"]

    #     assert all([
    #         expected_key in response.keys() for expected_key in expected_keys
    #     ])

    #     assert isinstance(response["status"], bool)

    #     assert isinstance(response["url"], str)

    #     assert isinstance(response["content"],
    #                       str) or response["content"] is None

    @pytest.mark.parametrize("module_name", MODULES_TO_BE_TESTED)
    def test_get(
        self,
        module_name: str,
        host_url: str,
        headers: Dict[str, str],
    ) -> None:

        api_url_base: str = f"{host_url}/{module_name}/cli-api"
        api_path: str = "public/status"

        api_url: str = f"{api_url_base}/{api_path}"

        response = get(url=api_url, headers=headers)

        assert isinstance(response, dict)

        expected_keys: List[str] = ["content", "status", "url"]

        assert all(
            [expected_key in response.keys() for expected_key in expected_keys])

        assert isinstance(response["status"], bool)

        assert isinstance(response["url"], str)

        assert isinstance(response["content"],
                          str) or response["content"] is None

    @pytest.mark.skip("Implement Test")
    def test_put(self) -> None:
        pass

    @pytest.mark.parametrize("module_name", MODULES_TO_BE_TESTED)
    def test_client_instantiation(
        self,
        client,
        module_name: str,
        test_configuration,
    ) -> None:
        assert client.auth_token is None
        assert test_configuration['api_token_name'] not in client.headers.keys()

    @pytest.mark.parametrize("module_name", MODULES_TO_BE_TESTED)
    def test_get_server_status(self, client, module_name: str) -> None:
        # We verify that the server is operational.
        server_status: str = client.get_server_status()
        expected_server_status: str = "Teselagen CLI API is operational."
        assert server_status == expected_server_status

    @pytest.mark.parametrize("module_name", MODULES_TO_BE_TESTED)
    def test_get_api_info_deauthorized(self, client, module_name: str) -> None:
        # The client should only be instantiated but not authorized.
        # with pytest.raises(AssertionError, match=r".*unauthorized.*"):
        api_info = client.get_api_info()
        assert "unauthorized" in api_info.lower()

    @pytest.mark.parametrize("module_name", MODULES_TO_BE_TESTED)
    def test_login(
        self,
        client,
        module_name: str,
        expiration_time: str,
        test_configuration,
    ) -> None:

        # LOGIN
        # We login the user with the CLI.
        client.login(  #username=credentials["test_user"],
            #passwd=credentials["test_password"],
            expiration_time=expiration_time)

        # We verify the client is authorized.
        api_info = client.get_api_info()
        assert "unauthorized" not in api_info.lower()

        # Now the token should be a string.
        assert isinstance(client.auth_token, str)

        # We verify that the API_TOKEN_NAME key has been added to the client
        # headers
        assert test_configuration["api_token_name"] in client.headers.keys()
        assert isinstance(
            client.headers[test_configuration["api_token_name"]],
            str,
        )

        # We get the current user (auth) information
        current_user = client.get_current_user()
        assert isinstance(current_user["content"]["username"], str)

        # LOGOUT
        # We logout the user from the CLI.
        client.logout(  #username=credentials["test_user"],
            #password=credentials["test_password"]
        )

        # We check the client is not authorized.
        api_info = client.get_api_info()
        assert "unauthorized" in api_info.lower()

    # @pytest.mark.skip(reason="This endpoint is currently not implemented")
    @pytest.mark.parametrize("module_name", MODULES_TO_BE_TESTED)
    def test_get_laboratories(self, logged_client: TeselaGenClient) -> None:
        client = logged_client

        # NOTE : Currently, laboratories information only exists in TEST,
        #        and its probably global.

        response: List[Dict[str, Any]] = client.get_laboratories()

        assert isinstance(response, list)
        assert len(response) > 0
        assert all(isinstance(element, dict) for element in response)
        assert all(key in element.keys()
                   for element in response
                   for key in ["id", "name"])

    @pytest.mark.parametrize("module_name", [MODULES_TO_BE_TESTED[0]])
    def test_select_lab_by_name(self, logged_client: TeselaGenClient) -> None:
        with patch.object(TeselaGenClient, 'get_laboratories') as get_lab_mock:
            labs = [{'id': 0, 'name': 'a'}, {'id': 1, 'name': 'b'}]
            get_lab_mock.return_value = labs
            client = logged_client
            client.select_laboratory(lab_name='b')
        assert int(client.headers['tg-active-lab-id']) == labs[1]['id']



#!/usr/bin/env python3
from typing import List, Optional
from dna_features_viewer import GraphicFeature, GraphicRecord


def show_crispr_grna_results(sequence:str, guides: List[dict], indexes: Optional[List[int]]=None, scoreField:str="onTargetScore"):
    """ Shows guide rnas results for CRISPR

    Args:
        sequence (str): A string containig the complete organism sequence
        guides (dict): A table on 'records' format that contains guides info. The
            required fields are `start` (int), `end` (ind), indicating the limits
            of the guide in sequence's index.
        indexes(list): Indexes (start and end) of the targeted sequence within the
            complete sequence. If not set, the targeting sequence is not shown.
        scoreField (str): Select which score from GRNA tool show in the chart. Available
            scores are "onTargetScore" (default) and "offTargetScore"
    """
    targeting_seq_feat = []
    # Show main targeted sequence if index are set. If not, we calculate indexes
    # to limit plot range at the `crop` instruction.
    if indexes is not None:
        targeting_seq_feat = [GraphicFeature(start=indexes[0], end=indexes[1], color="#cffccc", label="Sequence", strand=+1)]
    else:
        indexes = [min([x['start'] for x in guides]), max([x['end'] for x in guides])]
    # Plot records
    record = GraphicRecord(
        sequence = sequence,
        features = targeting_seq_feat +
                 [GraphicFeature(
                      start=x['start'],
                      end=x['end']+1,
                      color="#ffcccc",
                      label=f"{scoreField}: {x[scoreField]}",
                      strand=+1 if x['forward'] else -1) for x in guides])
    # Limit plot range
    record=record.crop((indexes[0]-10, indexes[1]+11)) # crop
    # Plot and set to show sequence
    ax, _ = record.plot(figure_width=20)
    record.plot_sequence(ax)


#!/usr/local/bin/python3
# Copyright (C) 2018 TeselaGen Biotechnology, Inc.
# License: MIT

from io import StringIO
import json
from os.path import join
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import pandas as pd
from teselagen.utils import DEFAULT_API_TOKEN_NAME
from teselagen.utils import DEFAULT_HOST_URL
from teselagen.utils import delete
from teselagen.utils import get
from teselagen.utils import post
from teselagen.utils import put

# NOTE : Related to Postman and Python requests
#       "body" goes into the "json" argument
#       "Query Params" goes into "params" argument


class TESTClient():

    def __init__(self, teselagen_client: Any):
        module_name: str = "test"

        self.host_url = teselagen_client.host_url
        self.headers = teselagen_client.headers
        # Here we define the Base CLI URL.
        api_url_base: str = f"{self.host_url}/{module_name}/cli-api"
        # Here we define the client endpoints
        # Example :
        #    self.some_endpoint_url: str = f"{api_url_base}/some_endpoint"

        # Assay Subjects
        self.get_assay_subjects_url: str = f"{api_url_base}/assay-subjects"
        self.get_assay_subject_url: str = join(api_url_base,
                                               "assay-subjects") + "/{}"
        self.create_assay_subjects_url: str = f"{api_url_base}/assay-subjects"
        self.delete_assay_subject_url: str = join(api_url_base,
                                                  "assay-subjects") + "/{}"
        self.put_assay_subject_descriptors_url: str = f"{api_url_base}/assay-subjects/descriptors"

        # Experiments
        self.get_experiments_url: str = f"{api_url_base}/experiments"
        self.create_experiment_url: str = f"{api_url_base}/experiments"
        self.delete_experiment_url: str = join(api_url_base,
                                               "experiments") + "/{}"

        # Assays
        self.get_assays_url: str = f"{api_url_base}/assays"
        self.get_assays_by_experiment_url: str = join(
            api_url_base, "experiments") + "/{}/assays"

        self.create_assay_url: str = join(api_url_base,
                                          "experiments") + "/{}/assays"
        self.delete_assay_url: str = join(api_url_base, "assays") + "/{}"

        self.post_assay_results_import_url: str = f"{api_url_base}/assays/results/import"
        self.get_assay_results_import_url: str = join(
            api_url_base, "assays") + "/results/import/{}"

        self.assay_results_url: str = join(api_url_base,
                                           "assays") + "/{}/results"

        # Files
        self.get_files_info_url: str = f"{api_url_base}/files"
        self.get_file_data_url: str = join(api_url_base, "files") + "/{}"
        self.delete_file_url: str = join(api_url_base, "files") + "/{}"
        self.upload_file_url: str = join(api_url_base, "files")
        self.upload_file_into_assay_url: str = join(api_url_base,
                                                    "assays") + "/{}/files"
        self.upload_file_into_experiment_url: str = join(
            api_url_base, "experiments") + "/{}/files"

        # Metadata
        self.get_metadata_url: str = join(api_url_base, "metadata") + "/{}"
        self.create_metadata_url: str = join(api_url_base, "metadata")
        self.delete_metadata_url: str = join(api_url_base,
                                             "metadata") + "/{}/{}"

    # Assay Subject Endpoints
    def create_assay_subject(self, name: str, assaySubjectClassId: int):
        body = [{"name": name, "assaySubjectClassId": str(assaySubjectClassId)}]

        response: Dict[str, Any] = post(url=self.create_assay_subjects_url,
                                        headers=self.headers,
                                        json=body)

        response["content"] = json.loads(response["content"])

        return response["content"]

    def get_assay_subjects(self,
                           assay_subject_ids: Optional[Union[int,
                                                             List[int]]] = None,
                           summarized: bool = True):
        """
        This function fetches one or many assay subject records from TEST. It receives either an integer ID or a list of integer IDs
        through the 'assay_subject_ids' argument, which correspond to the the assay subject IDs.

        Args:
            assay_subject_ids(Optional[Union[int, List[int]]]): Either an integer, a list of integers or None. When integers are passed,
                these are treated as the assay subject IDs used to query the database, if None, all assay subjects are returned.

            summarized (bool): Flag indicating whether the returned assay subject records should be summarized or if full assay subject objects
                should be returned. Default is True.

        Returns:
            - A list of assay subject records (summarized or fully detailed). Depending on the summarized parameter each property in the list is listed below:

            Assay Subject record structure:
                - id (str): ID of the Assay Subject (summarized and full).
                - name (str): Name of the Assay Subject (summarized and full).
                - assaySubjectClass (dict): A JSON with assay subject class information (summarized and full).
                - descriptors (List[dict]): A list of JSON records with the assay subject descriptors information (full).
                - assaySubjectGroups (List[dict]): A list of JSON records with the assay subject groups information (full).
                - experiments (List[dict]): A list of JSON records with the assay subject experiments information (full).
                - assays (List[dict]): A list of JSON records with the assay subject assays information (full).
        """

        url = ""
        params: Dict[str, Any] = {"summarized": str(summarized).lower()}
        if isinstance(assay_subject_ids, list):
            url = self.get_assay_subjects_url
            params["ids[]"] = assay_subject_ids
        elif isinstance(assay_subject_ids, int):
            url = self.get_assay_subject_url.format(assay_subject_ids)
        elif assay_subject_ids is None:
            url = self.get_assay_subjects_url
        else:
            raise TypeError(
                f"Argument 'assay_subject_ids' must of type int or List[int]. Not type: {type(assay_subject_ids)}"
            )
        response: Dict[str, Any] = get(
            url=url,
            params=params,
            headers=self.headers,
        )

        # response["content"] = [{"id" : str, "name": str}, ...]
        response["content"] = json.loads(response["content"])

        return response["content"]

    def delete_assay_subjects(self, assay_subject_ids: Union[int, List[int]]):
        """
        This function deletes one or many assay subject records from TEST. It receives an int ID or a list of int IDs
        through the 'assay_subject_ids' argument, which correspond to the the assay subject IDs.
        """
        params = {}
        if isinstance(assay_subject_ids, list):
            params["ids[]"] = assay_subject_ids
        elif isinstance(assay_subject_ids, int):
            params["ids[]"] = [assay_subject_ids]
        else:
            raise TypeError(
                f"Argument 'assay_subject_ids' must of type int or List[int]. Not type: {type(assay_subject_ids)}"
            )

        response = delete(
            url=self.delete_assay_subject_url.format(""),
            params=params,
            headers=self.headers,
        )

        return response["content"]

    def put_assay_subject_descriptors(
        self,
        mapper: List[dict],
        file_id: Optional[int] = None,
        filepath: Optional[str] = None,
        createSubjectsFromFile: Optional[bool] = False,
    ):
        """
            Calls Teselagen TEST API endpoint: `PUT /assay-subjects/descriptors`.
            The data can be passed via a local filepath or either the file ID after already uploading it.

            Args:
                mapper (List[dict]): This is the JSON mapper used by the endpoint to understand each of the file columns. This mapper
                    should be a list of Python Dictionary representing each structured header with a 'name', 'class' and 'subClassId' key.
                    For more information on the mappers structure refer to https://api-docs.teselagen.com/#operation/SubjectsPutAssaySubjectDecriptors
                file_id (Optional[int]) : File identifier.
                filepath (Optional[str]) : Local location of the file.
                createSubjectsFromFile (bool) : Flag that indicates whether to create new Assay Subject found in the file.

            Returns: a JSON object with a success status, the number of results inserted, and whether new assay subjects
                were created during the insert.
        """
        # Implements the ability to do the file upload behind the scenes.
        if (file_id is None):
            if filepath is not None and (Path(filepath).exists()):
                file = self.upload_file(filepath=filepath)
                file_id = file['id']
            else:
                raise FileNotFoundError(f"File: {filepath} not found")

        body = {
            "fileId": file_id,
            "mapper": mapper,
            "createSubjectsFromFile": createSubjectsFromFile
        }

        response: Dict[str, Any] = put(
            url=self.put_assay_subject_descriptors_url,
            headers=self.headers,
            json=body,
        )

        response["content"] = json.loads(response["content"])

        return response["content"]

    def import_assay_subject_descriptors(
        self,
        mapper: List[dict],
        file_id: Optional[int] = None,
        filepath: Optional[str] = None,
        createSubjectsFromFile: Optional[bool] = False,
    ):
        #TODO: This is a temporary implementation while a better solution
        # for long task is implemented.

        # Implements the ability to do the file upload behind the scenes.
        if (file_id is None):
            if filepath is not None and (Path(filepath).exists()):
                file = self.upload_file(filepath=filepath)
                file_id = file['id']
            else:
                raise FileNotFoundError(f"File: {filepath} not found")
        body = {
            "fileId": file_id,
            "mapper": mapper,
            "createSubjectsFromFile": createSubjectsFromFile
        }
        response: Dict[str, Any] = post(
            url=self.post_assay_results_import_url,
            headers=self.headers,
            json=body,
        )
        parsed_content = json.loads(response["content"])
        if 'message' in parsed_content.keys():
            parsed_content['message'] = parsed_content['message'].replace(
                "Assay results", "Assay Subject descriptor")
        response["content"] = parsed_content
        return response["content"]

    def get_assay_subjects_descriptor_import_status(
        self,
        importId: str,
    ) -> Any:
        '''
        Calls Teselagen TEST API endpoint: `GET /assays/results/import/:importId`.
        Args:
            - importId (string): ID of an assay result import process.
        Returns: a JSON object with information on the status of an assay result import job.
        '''
        try:
            response: Dict[str, Any] = get(
                url=self.get_assay_results_import_url.format(importId),
                headers=self.headers,
            )
        except Exception as e:
            # TODO : Use a logger
            print("Error:", e)
            return None
        response["content"] = json.loads(response["content"])
        response.pop('url')
        return response

    # Experiments Endpoints

    def get_experiments(self) -> List[Dict[str, Any]]:
        """

        Fetches all experiments from the Laboratory selected with the select_laboratory function.

        Args :

        Returns :
            () : A list of experiments objects.

        ```
            [
                {"id": "1", "name": "Experiment 1"},
                {"id": "2", "name": "Experiment 2"}
            ]
        ```

        """

        response: Dict[str, Any] = get(
            url=self.get_experiments_url,
            headers=self.headers,
        )

        # response["content"] = [{"id" : str, "name": str}, ...]
        response["content"] = json.loads(response["content"])

        return response["content"]

    def create_experiment(self, experiment_name: str) -> Dict[str, Any]:

        experiment: List[Any] = []
        exp_response: Dict[str, Any] = {}
        experiments = self.get_experiments()
        experiment = list(
            filter(lambda x: x['name'] == experiment_name, experiments))
        if len(experiment) != 1:
            body = {"name": experiment_name}
            response: Dict[str, Any] = post(url=self.create_experiment_url,
                                            headers=self.headers,
                                            json=body)
            exp_response = json.loads(response["content"])[0]
            # Now we GET experiments from db in order to return
            # the complete experiment to the user
            experiment = list(
                filter(
                    lambda x: x['id'] == exp_response['id'],
                    self.get_experiments(),
                ))
        if len(experiment) == 0:
            raise IOError(
                f"Error while looking for new id {exp_response['id']}")

        return experiment[0]

    def delete_experiment(self, experiment_id: int) -> None:
        """ Deletes an experiment with ID=`experiment_id`. """
        response = delete(
            url=self.delete_experiment_url.format(experiment_id),
            headers=self.headers,
        )

        return None

    # Assay Endpoints

    def get_assays(
        self,
        experiment_id: Optional[str] = None,
    ) -> List[Dict[str, Any]]:
        """
            Fetches all assays from the experiment specified in `experiment_id`.
            If no `experiment_id` is passed, all assays from the selected
            Laboratory are returned.

            Args :

                experiment_id (int):
                    Experiment identifier.

            Returns :
                (List[Dict[str, Any]]) :
                    A list of assays objects.

            ```
                [
                    {
                        "id"         : "1",
                        "name"       : "Assay 1",
                        "experiment" : {
                                            "id"   : "1",
                                            "name" : "Experiment 1"
                                        }
                    },
                    {
                        "id"         : "2",
                        "name"       : "Assay 2",
                        "experiment" : {
                                            "id"   : "1",
                                            "name" : "Experiment 1"
                                        }
                    },
                ]
            ```
        """

        response: Dict[str, Any] = get(
            url=self.get_assays_by_experiment_url.format(experiment_id)
            if experiment_id else self.get_assays_url,
            headers=self.headers,
        )

        response["content"] = json.loads(response["content"])

        return response["content"]

    def create_assay(
        self,
        experiment_id: str,
        assay_name: str,
        parser_id: Optional[int] = None,
    ) -> Dict[str, Any]:

        body = {
            "name": assay_name,
            "parserId": str(parser_id) if parser_id else None
        }

        try:
            response: Dict[str, Any] = post(
                url=self.create_assay_url.format(experiment_id),
                headers=self.headers,
                json=body,
            )
        except Exception as e:
            # TODO : Use a logger
            raise Exception(e)

        assay_res = json.loads(response["content"])[0]
        # Retrieve the created object
        assay = list(
            filter(
                lambda x: x['id'] == assay_res['id'],
                self.get_assays(experiment_id=experiment_id),
            ))
        if len(assay) == 0:
            raise IOError(f"Can't find new id {assay_res['id']}")
        return assay[0]

    def delete_assay(self, assay_id: str) -> Any:
        """ Deletes an Assay with ID=`assay_id`. """
        response: Dict[str, Any] = delete(
            url=self.delete_assay_url.format(assay_id),
            headers=self.headers,
        )

        content = json.loads(response['content'])

        return content

    def delete_assays(self, assay_ids: List[str]) -> List[Any]:
        """ Deletes assays referenced by the IDs in the assay_ids list. """
        response: List[Any] = []
        for assay_id in assay_ids:
            res = self.delete_assay(assay_id=assay_id)
            response.append(res)

        return response

    def put_assay_results(
        self,
        mapper: List[dict],
        assay_id: Optional[Union[int, str]] = None,
        file_id: Optional[Union[int, str]] = None,
        filepath: Optional[Union[str, Path]] = None,
        assay_name: Optional[str] = None,
        experiment_id: Optional[Union[str, int]] = None,
        createSubjectsFromFile: Optional[bool] = True,
        createMeasurementTargetsFromFile: Optional[bool] = True,
    ):
        """
            Calls Teselagen TEST API endpoint: `PUT /assays/:assayId/results`.
            The data can be passed via a local filepath or either the file ID after already uploading it.

            Args:
                mapper (List[dict]): This is the JSON mapper used by the endpoint to understand each of the file columns. This mapper
                    should be a list of Python Dictionary representing each structured header with a 'name', 'class' and 'subClassId' key.
                    For more information on the mappers structure refer to https://api-docs.teselagen.com/#operation/AssaysPutAssayResults
                assay_id (int) : Assay identifier.
                file_id (int) : File identifier.
                filepath (int) : Local location of the file.
                assay_name (str) : Name of the assay into which insert the assay results.
                experiment_id (number) : Experiment identifier. Only used when passed and 'assay_name' an no 'assay_id'.
                createSubjectsFromFile (bool) : Flag that indicates whether to create new Assay Subject found in the file.
                createMeasurementTargetsFromFile (bool) : Flag that indicates whether to create new Measurement Target metadata records found in the file.

            Returns: a JSON object with a success status, the number of results inserted, and whether new assay subjects and/or mesurement targets
                were created during the insert.
        """
        if (assay_id is None and assay_name is not None):
            # Supports creating a new assay by providing an assay name and an experiment ID.
            assay_id = self.get_or_create_assay(
                assay_name=assay_name,
                experiment_id=str(experiment_id),
            )
        else:
            raise Exception(
                f"Please provide a valid 'assay_id' or 'assay_name'.")
        # Implements the ability to do the file upload behind the scenes.
        if (file_id is None and filepath is not None):
            file_id = self.get_or_upload_file(
                filepath=filepath,
                assay_id=str(assay_id),
            )
        else:
            raise Exception(
                f"Please provide a valid 'file_id' or an existant 'filepath'.")

        body = {
            "assayId": assay_id,
            "fileId": file_id,
            "mapper": mapper,
            "createSubjectsFromFile": createSubjectsFromFile,
            "createMeasurementTargetsFromFile": createMeasurementTargetsFromFile
        }
        try:
            response: Dict[str, Any] = put(
                url=self.assay_results_url.format(assay_id),
                headers=self.headers,
                json=body,
            )
            print(f"response: {response}")
        except Exception as e:
            # TODO : Use a logger
            print("Error:", e)
            return None
        response["content"] = json.loads(response["content"])

        return response["content"]

    def import_assay_results(
        self,
        mapper: List[dict],
        assay_id: Optional[Union[int, str]] = None,
        file_id: Optional[Union[int, str]] = None,
        filepath: Optional[Union[str, Path]] = None,
        assay_name: Optional[str] = None,
        experiment_id: Optional[Union[str, int]] = None,
    ):
        """
            Calls Teselagen TEST API endpoint: `POST /assays/results/importer`.
            The data can be passed via a local filepath or either the file ID after already uploading it.

            Args:
                mapper (List[dict]): This is the JSON mapper used by the endpoint to understand each of the file columns. This mapper
                    should be a list of Python Dictionary representing each structured header with a 'name', 'class' and 'subClassId' key.
                    For more information on the mappers structure refer to https://api-docs.teselagen.com/#operation/AssaysPutAssayResults
                assay_id (int) : Assay identifier.
                file_id (int) : File identifier.
                filepath (int) : Local location of the file.
                assay_name (str) : Name of the assay into which insert the assay results.
                experiment_id (number) : Experiment identifier. Only used when passed and 'assay_name' an no 'assay_id'.

            Returns: a JSON object with a status and an import process ID. Which can be used to check the status of the import progress
                by means of the 'get_assay_results_import_status' function.
        """
        if (assay_id is None and assay_name is not None):
            # Supports creating a new assay by providing an assay name and an experiment ID.
            assay_id = self.get_or_create_assay(
                assay_name=assay_name,
                experiment_id=str(experiment_id),
            )
        else:
            raise Exception(
                f"Please provide a valid 'assay_id' or 'assay_name'.")
        # Implements the ability to do the file upload behind the scenes.
        if (file_id is None and filepath is not None):
            file_id = self.get_or_upload_file(
                filepath=filepath,
                assay_id=str(assay_id),
            )
        else:
            raise Exception(
                f"Please provide a valid 'file_id' or an existant 'filepath'.")
        body = {
            "assayId": assay_id,
            "fileId": file_id,
            "mapper": mapper,
        }
        try:
            response: Dict[str, Any] = post(
                url=self.post_assay_results_import_url,
                headers=self.headers,
                json=body,
            )
        except Exception as e:
            # TODO : Use a logger
            print("Error:", e)
            return None
        response["content"] = json.loads(response["content"])

        return response["content"]

    def get_assay_results_import_status(
        self,
        importId: str,
    ) -> Any:
        '''
        Calls Teselagen TEST API endpoint: `GET /assays/results/import/:importId`.

        Args:
            - importId (string): ID of an assay result import process.

        Returns: a JSON object with information on the status of an assay result import job.
        '''
        try:
            response: Dict[str, Any] = get(
                url=self.get_assay_results_import_url.format(importId),
                headers=self.headers,
            )
        except Exception as e:
            # TODO : Use a logger
            print("Error:", e)
            return None
        response["content"] = json.loads(response["content"])

        return response

    def get_assay_results(
        self,
        assay_id: int,
        as_dataframe: bool = True,
        with_subject_data: bool = True,
        group: bool = True,
    ):
        """
            Calls Teselagen TEST API endpoint: `GET /assays/:assayId/results`.
            More information at https://api-docs.teselagen.com/#operation/AssaysGetAssayResults.

            Args:
                assay_id (int): Assay identifier.
                as_dataframe (bool): Flag indicating whether to return the data as a dataframe (default=True).
                with_subject_data (bool): Flag indicating whether to return the assay results together with a more complete
                    information on the assay subjects (default=True).
                group (bool): Flag indicating whether to group the assay results and assay subjects by the corresponding tabular indexes.
                    Only used when the 'as_dataframe' is set to True (default=True).

            Returns: Either a dataframe or a JSON Array. The information included in the returned object is all the assay results,
                plus the assay name, and assay subject information (if 'with_subject_data' is set to True).
        """
        # NOTE: depending on the different flags, the order of the columns may vary.
        api_result = self._get_assay_results_from_api(assay_id=assay_id)

        assay_results = api_result['results']
        tabular_assay_results, assay_result_indexes =\
            self._tabular_format_assay_result_data(assay_results)

        if as_dataframe:
            final_results = pd.DataFrame(tabular_assay_results).set_index(
                assay_result_indexes[0])
            final_results.insert(0, "Assay", api_result['name'])
            # If required, group by the assay results and assay subject indexes.
            # Usually these indexes are going to be the assay subject id and any reference dimension found in the assay results.
            final_results = final_results.groupby(
                by=[*assay_result_indexes
                   ]).first().reset_index() if group else final_results

            if with_subject_data:
                assaySubjectIds = list(
                    set([
                        assay_result['assaySubjectId']
                        for assay_result in assay_results
                    ]))
                # assay_subjects = [assaySubject for assaySubject in tqdm(self.get_assay_subjects(assaySubjectIds))]
                assay_subjects = self.get_assay_subjects(
                    assay_subject_ids=assaySubjectIds, summarized=False)

                tabular_assay_subjects, assay_subject_indexes =\
                    self._tabular_format_assay_subject_data(assay_subjects)

                assay_subjects_df = pd.DataFrame(
                    tabular_assay_subjects).set_index(assay_subject_indexes)

                # Here we merge both dataframes.
                final_results = assay_subjects_df.merge(
                    final_results,
                    left_on=assay_subject_indexes,
                    right_on=assay_subject_indexes,
                )

        else:
            if with_subject_data:
                assaySubjectIds = list(
                    set([
                        assay_result['assaySubjectId']
                        for assay_result in assay_results
                    ]))
                # assay_subjects = [assaySubject for assaySubject in tqdm(self.get_assay_subjects(assaySubjectIds))]
                assay_subjects = self.get_assay_subjects(
                    assay_subject_ids=assaySubjectIds,
                    summarized=False,
                )
                tabular_assay_subjects, assay_subject_indexes =\
                    self._tabular_format_assay_subject_data(assay_subjects)

                final_results = [{
                    **{
                        "Assay": api_result['name']
                    },
                    **assay_subject,
                    **assay_result
                } for (assay_subject, assay_result
                      ) in zip(tabular_assay_subjects, tabular_assay_results)]
            else:
                final_results = [{
                    **{
                        "Assay": api_result['name']
                    },
                    **assay_result
                } for assay_result in tabular_assay_results]
        return final_results

    def _get_assay_results_from_api(self, assay_id: int):

        url = self.assay_results_url.format(assay_id)

        response: Dict[str, Any] = get(url=url, headers=self.headers)

        api_result = json.loads(response["content"])

        return api_result

    # File Endpoints

    def get_files_info(self) -> List[Dict[str, Any]]:
        """
            Fetches all files from the selected Laboratory.

            Returns :
                () :
                    A list of assays objects.

            ```
                [{
                    "id": "1",
                    "name": "File 1",
                    "assay": {
                        "id": "1",
                        "name": "Assay 1"
                    },
                    "experiment": {...}
                },
                {
                    "id": "2",
                    "name": "File 2",
                    "assay": null,
                    "experiment": {...}
                }]
            ```
        """

        response: Dict[str, Any] = get(
            url=self.get_files_info_url,
            headers=self.headers,
        )

        response["content"] = json.loads(response["content"])

        return response["content"]

    def upload_file(
        self,
        filepath: Union[str, Path],
        experiment_id: Optional[int] = None,
        assay_id: Optional[str] = None,
    ):
        """
            Uploads a file. The request body is of type "multipart/form-data".

            It requires the "filepath" and optionally with an "assay_id".

            If no `assay_id` is passed the file will be uploaded linked to no
            assay.

            NB: If an assay_id with an assigned parser is passed the file will be
            automatically parsed with such parser.

            Args :
                filepath (str) :
                    Path to the file to be uploaded.

                experiment_id (Optional[int]) :
                    Experiment identifier.

                assay_id (Optional[int]) :
                    Assay identifier.

            Returns :
        """

        filepath = Path(filepath)
        multipart_form_data = {
            'file': (filepath.name, open(filepath, 'rb')),
        }

        # We need a header file wihtout the 'Content-Type' key because this is a 'multipart/form-data' request
        # unlike the others which have Content-Type = 'application/json'. Here, only the authorization token is needed.
        headers = self.headers.copy()
        del headers['Content-Type']

        upload_file_url = self.upload_file_into_assay_url.format(
            assay_id
        ) if assay_id else self.upload_file_into_experiment_url.format(
            experiment_id) if experiment_id else self.upload_file_url
        response: Dict[str, Any] = post(
            url=upload_file_url,
            headers=headers,
            files=multipart_form_data,
        )
        res_files_info = json.loads(response["content"])
        if not isinstance(res_files_info, dict):
            raise IOError(
                f"There was a problem with upload (maybe check assay_id): response: {response}"
            )
        # Build our object to be returned (new file_info)
        # Get the file info with the right id
        files_info = list(
            filter(
                lambda x: x['id'] == res_files_info['id'],
                self.get_files_info(),
            ))
        if len(files_info) == 0:
            raise IOError(
                f"Name {multipart_form_data['file'][0]} not found in uploaded files"
            )
        return files_info[0]

    def download_file(self, file_id: Union[int, str]) -> StringIO:
        """
            It will return the data contents of the corresponding file with ID
            specified in the "file_id" argument.

            It returns a StringIO object, which keeps the file data serialized.
            One could take this serialized data and write it into a file or directly rad it with pandas.

            Args:
                file_id (int) :
                        File identifier.

            Returns:
                a StringIO object with the data.
        """

        response: Dict[str, Any] = get(
            url=self.get_file_data_url.format(file_id),
            headers=self.headers,
        )

        return StringIO(response["content"])

    def delete_file(self, file_id: Union[int, str]) -> Dict[str, Any]:
        """ Deletes a File with ID=`file_id`. """
        response: Dict[str, Any] = delete(
            url=self.delete_file_url.format(file_id),
            headers=self.headers,
        )

        content = json.loads(response["content"])

        return content

    # Metadata Endpoints

    def get_metadata(self, metadataType: str, metadataTypeFields: str = None):
        """

        Returns metadata records according to the 'metaDataType' path parameter. Available metaDataTypes are:

            - assaySubjectClass
            - measurementTarget
            - measurementType
            - unit
            - unitScale
            - unitDimension
            - descriptorType

        Args :

            metadataType (str): The type of a metadata. Must be one of the available metadata types listed above.

        Returns : A JSON object with the metadata records belonging to the requested metadata type.

        ```
            [
                {"id": "1", "name": "Metadata Record 1"},
                {"id": "2", "name": "Metadata Record 2"},
            ]
        ```

        """

        response: Dict[str, Any] = get(
            url=self.get_metadata_url.format(metadataType),
            headers=self.headers,
        )

        response["content"] = json.loads(response["content"])

        return response["content"]

    def create_metadata(
        self,
        metadataType: str,
        metadataRecord: Union[List[dict], dict],
    ):
        """
            Calls Teselagen TEST API endpoint: `POST /metadata`.
            More information at https://api-docs.teselagen.com/#operation/MetadataCreateMetadata.

            Args:
                metadataType (str): Name of the metadata type/class.
                metadataRecord (Union[List[dict], dict]): Either an array of metadata records or a single one.
                    These should follow the required structure of a metadata record. For more information on this
                    refer to the above API documentation link.
        """
        body = {"metaData": {metadataType: metadataRecord}}

        response: Dict[str, Any] = post(url=self.create_metadata_url,
                                        headers=self.headers,
                                        json=body)

        # [{ id: "3" }]
        response["content"] = json.loads(response["content"])

        return response["content"]

    def delete_metadata(self, metadataType: str, metadataId: int):
        response = delete(
            url=self.delete_metadata_url.format(
                metadataType,
                metadataId,
            ),
            headers=self.headers,
        )

        # response["content"] = json.loads(response["content"])

        return True

    # Others

    # TEST Client Utils

    def get_or_create_assay(
        self,
        assay_name: str,
        experiment_id: str,
    ) -> Optional[str]:
        ''' Supports creating a new assay by providing an assay name and an experiment ID.'''
        assay_id = None
        if (experiment_id is not None):
            assays = self.get_assays()
            assay = list(
                filter(
                    lambda x: x['name'] == assay_name and x['experiment']['id']
                    == experiment_id, assays))
            assay_id = assay[0]['id'] if len(assay) > 0 else self.create_assay(
                experiment_id=experiment_id, assay_name=assay_name)['id']
        else:
            raise Exception(f"Please provide a valid 'experiment_id'.")

        return assay_id

    def get_or_upload_file(
        self,
        filepath: Union[str, Path],
        assay_id: str,
    ) -> Optional[str]:
        file_id = None
        if Path(filepath).exists():
            # See the current files already uploaded to the assay.
            files = self.get_files_info()
            assay_files = list(
                filter(
                    lambda x: x['assay'] is not None and x['assay']['id'] ==
                    assay_id and filepath is not None and Path(x[
                        'name']).name == Path(filepath).name,
                    files,
                ))

            if (len(assay_files) > 0):
                file_id = assay_files[0]['id']
            # NOTE: When a file with the same name has already been uploaded into the Assay, do not upload the file again.
            else:
                file = self.upload_file(
                    filepath=filepath,
                    assay_id=assay_id,
                )
                file_id = file['id']
        else:
            raise FileNotFoundError(f"Prvided 'filepath' does not exist.")
        return file_id

    def _tabular_format_assay_subject_data(self, assay_subjects_data: Any):
        tabular_assay_subjects = []
        for assay_subject_data in assay_subjects_data:
            assay_subject_row_dict = {
                "Subject ID": assay_subject_data['id'],
                "Subject Name": assay_subject_data['name'],
                "Subject Class": assay_subject_data['assaySubjectClass']['name']
            }
            for descriptor in assay_subject_data['descriptors']:
                assay_subject_row_dict[descriptor['descriptorType']
                                       ['name']] = descriptor['value']

            tabular_assay_subjects.append(assay_subject_row_dict)

        indexes = ["Subject ID"]
        return tabular_assay_subjects, indexes

    def _tabular_format_assay_result_data(self, assay_result_data: Any):
        tabular_assay_results = []
        assaySubjectColumnName = 'Subject ID'
        assaySubjectIds = set()
        referenceDimensions = set()
        measurementTypes = set()
        for result in assay_result_data:
            # The assay subject ID is important because a tabular form would be indexed by these.
            assaySubjectId = result['assaySubjectId']
            assaySubjectIds.add(assaySubjectId)
            tabular_row_assay_result_dict = {
                assaySubjectColumnName: assaySubjectId
            }

            # reference dimensions are important when formatting assay results, because a tabular form
            # would be indexed by these.
            if "reference" in result:
                referenceDimension = f"{result['reference']['name']}_({result['reference']['unit']})"
                referenceDimensions.add(referenceDimension)
                tabular_row_assay_result_dict[referenceDimension] = result[
                    'reference']['value']

            # These do not need specal index treatment for a tabular form.
            # However these are still collected and returned in case of need.
            measurementType = f"{result['result']['name']}_({result['result']['unit']})"
            measurementTypes.add(measurementType)
            tabular_row_assay_result_dict[measurementType] = result['result'][
                'value']

            tabular_assay_results.append(tabular_row_assay_result_dict)

        if len(referenceDimensions) > 1:
            # TODO: add support for multiple reference dimensions.
            raise Exception("Multiple Reference Dimensions not supported.")

        indexes = [assaySubjectColumnName, *list(referenceDimensions)]

        return tabular_assay_results, indexes
